# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import random
import uuid
from collections.abc import Iterable
from typing import Any
from uuid import UUID

import polars as pl

from pydiverse.common import (
    Dtype,
    Int,
    String,
)
from pydiverse.transform._internal.backend.table_impl import (
    TableImpl,
    get_left_right_on,
    split_join_cond,
)
from pydiverse.transform._internal.backend.targets import Pandas, Polars, Target
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.ops.op import Ftype
from pydiverse.transform._internal.tree import types, verbs
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import (
    CaseExpr,
    Cast,
    Col,
    ColExpr,
    ColFn,
    EvalAligned,
    LiteralCol,
    Order,
    Series,
)


class PolarsImpl(TableImpl):
    backend_name = "polars"

    def __init__(self, name: str, df: pl.DataFrame | pl.LazyFrame):
        self.df = df if isinstance(df, pl.LazyFrame) else df.lazy()
        self.df = self.df.cast(
            {
                pl.Datetime("ns"): pl.Datetime("us"),
                pl.Datetime("ms"): pl.Datetime("us"),
            }
        )
        super().__init__(
            name,
            {name: Dtype.from_polars(pl_type) for name, pl_type in df.collect_schema().items()},
        )

    def _table_def_repr(self) -> str:
        return f"Table(df, Polars(), name='{self.name}')"

    @staticmethod
    def build_query(nd: AstNode) -> None:
        return None

    @staticmethod
    def export(
        nd: AstNode,
        target: Target,
        *,
        schema_overrides: dict[UUID, Any],  # TODO: use this
    ) -> Any:
        lf, name_in_df, select, _ = compile_ast(nd)
        lf = lf.select(*(name_in_df[uid] for uid in select))
        assert isinstance(lf, pl.LazyFrame)

        if isinstance(target, Polars):
            if not target.lazy:
                lf = lf.collect()
            lf.name = nd.name
            return lf

        elif isinstance(target, Pandas):
            return lf.collect().to_pandas(use_pyarrow_extension_array=True)

        raise AssertionError

    def _clone(self) -> tuple["PolarsImpl", dict[AstNode, AstNode], dict[UUID, UUID]]:
        cloned = PolarsImpl(self.name, self.df.clone())
        return (
            cloned,
            {self: cloned},
            {self.cols[name]._uuid: cloned.cols[name]._uuid for name in self.cols.keys()},
        )


# merges descending and null_last markers into the ordering expression
def merge_desc_nulls_last(
    order_by: list[pl.Expr], descending: list[bool], nulls_last: list[bool | None]
) -> list[pl.Expr]:
    merged = []
    for ord, desc, nl in zip(order_by, descending, nulls_last, strict=True):
        # try to avoid this workaround whenever possible
        if nl is not None or desc:
            numeric = ord.rank("dense").cast(pl.Int64)
            if desc:
                numeric = -numeric
            if nl is True:
                numeric = numeric.fill_null(pl.len().cast(pl.Int64) + 1)
            elif nl is False:
                numeric = numeric.fill_null(-pl.len().cast(pl.Int64) - 1)
            merged.append(numeric)
        else:
            merged.append(ord)

    return merged


def compile_order(order: Order, name_in_df: dict[UUID, str]) -> tuple[pl.Expr, bool, bool | None]:
    return (
        compile_col_expr(order.order_by, name_in_df),
        order.descending,
        order.nulls_last,
    )


def compile_col_expr(
    expr: ColExpr,
    name_in_df: dict[UUID, str],
    op_kwargs: dict[str, Any] | None = None,
) -> pl.Expr:
    if isinstance(expr, Col):
        if expr._uuid in name_in_df:
            return pl.col(name_in_df[expr._uuid])
        return expr.export(Polars)

    elif isinstance(expr, ColFn):
        impl = PolarsImpl.get_impl(expr.op, tuple(arg.dtype() for arg in expr.args))
        args: list[pl.Expr] = [compile_col_expr(arg, name_in_df, op_kwargs=op_kwargs) for arg in expr.args]

        if (partition_by := expr.context_kwargs.get("partition_by")) is not None:
            partition_by = [compile_col_expr(pb, name_in_df, op_kwargs=op_kwargs) for pb in partition_by]

        arrange = expr.context_kwargs.get("arrange")
        if arrange:
            order_by, descending, nulls_last = zip(
                *[compile_order(order, name_in_df) for order in arrange], strict=True
            )

        # The following `if` block is absolutely unnecessary and just an optimization.
        # Otherwise, `over` would be used for sorting, but we cannot pass descending /
        # nulls_last there and the required workaround is probably slower than polars`s
        # native `sort_by`.
        if arrange and not partition_by and len(args) > 0:
            # order the args. if the table is grouped by group_by or
            # partition_by=, the groups will be sorted via over(order_by=)
            # anyways so it need not be done here.
            args[0] = args[0].sort_by(
                by=order_by,
                descending=descending,
                nulls_last=[nl if nl is not None else False for nl in nulls_last],
            )

        if expr.op in (ops.rank, ops.dense_rank):
            assert len(expr.args) == 0
            args = [pl.struct(merge_desc_nulls_last(order_by, descending, nulls_last))]
            arrange = None

        value: pl.Expr = impl(*args, _partition_by=partition_by, **(op_kwargs or {}))

        # TODO: currently, count and list_agg are the only aggregation function where we
        # don't want to return null for cols containing only null values. If this
        # happens for more aggregation functions, make this configurable in e.g. the
        # operator spec.     --> do we want it for str.join??
        if expr.op.ftype == Ftype.AGGREGATE and expr.op not in (
            ops.count_star,
            ops.list_agg,
        ):
            # In `sum` / `any` and other aggregation functions, polars puts a
            # default value (e.g. 0, False) for empty columns, but we want to put
            # Null in this case to let the user decide about the default value via
            # `fill_null` if he likes to set one.
            assert all(types.is_const(arg.dtype()) for arg in expr.args[1:])
            value = pl.when(args[0].count() == 0).then(None).otherwise(value)

        if partition_by:
            # when doing sort_by -> over in polars, for whatever reason the
            # `nulls_last` argument is ignored. thus when both a grouping and an
            # arrangement are specified, we manually add the descending and
            # nulls_last markers to the ordering.
            if arrange:
                order_by = merge_desc_nulls_last(order_by, descending, nulls_last)
            else:
                order_by = None
            value = value.over(partition_by, order_by=order_by)

        elif arrange and expr.op.ftype == Ftype.WINDOW:
            # the function was executed on the ordered arguments. here we
            # restore the original order of the table.
            # TODO: use arg_sort?
            inv_permutation = pl.int_range(0, pl.len(), dtype=pl.Int64()).sort_by(
                by=order_by,
                descending=descending,
                nulls_last=[nl if nl is not None else False for nl in nulls_last],
            )
            value = value.sort_by(inv_permutation)

        return value

    elif isinstance(expr, CaseExpr):
        assert len(expr.cases) >= 1
        compiled = pl  # to initialize the when/then-chain
        for cond, val in expr.cases:
            compiled = compiled.when(compile_col_expr(cond, name_in_df, op_kwargs=op_kwargs)).then(
                compile_col_expr(val, name_in_df, op_kwargs=op_kwargs)
            )
        if expr.default_val is not None:
            compiled = compiled.otherwise(compile_col_expr(expr.default_val, name_in_df, op_kwargs=op_kwargs))
        return compiled

    elif isinstance(expr, LiteralCol):
        return pl.lit(
            expr.val,
            # only give the type explicitly if we can still be sure about it
            # in nested lists with ints / floats mixed we do not give guarantees
            dtype=expr.dtype().to_polars() if types.is_subtype(expr.dtype()) else None,
        )

    elif isinstance(expr, Cast):
        if (expr.target_type.is_int() or expr.target_type.is_float()) and types.without_const(
            expr.val.dtype()
        ) == String():
            expr.val = expr.val.str.strip()

        compiled = compile_col_expr(expr.val, name_in_df, op_kwargs=op_kwargs)
        compiled = compiled.cast(expr.target_type.to_polars(), strict=expr.strict)

        if types.without_const(expr.val.dtype()).is_float() and expr.target_type == String():
            compiled = compiled.replace("NaN", "nan")

        return compiled

    elif isinstance(expr, EvalAligned):
        return compile_col_expr(expr.val, name_in_df, op_kwargs)

    elif isinstance(expr, Series):
        return expr.val

    raise AssertionError


def rename_overwritten_cols(
    new_names: Iterable[str],
    df: pl.LazyFrame,
    name_in_df: dict[UUID, str],
    *,
    names_to_consider: set[UUID] | None = None,
) -> tuple[pl.LazyFrame, dict[UUID, str]]:
    if names_to_consider is None:
        names_to_consider = set(name_in_df.values())
    overwritten = names_to_consider.intersection(new_names)

    if overwritten:
        name_map = {name: f"{name}:{str(hex(uuid.uuid1().int))[2:]}" for name in overwritten}
        name_in_df = {uid: (name_map[name] if name in name_map else name) for uid, name in name_in_df.items()}
        df = df.rename(name_map)

    return df, name_in_df


def compile_ast(
    nd: AstNode,
) -> tuple[pl.LazyFrame, dict[UUID, str], list[UUID], list[UUID]]:
    if isinstance(nd, verbs.Verb):
        df, name_in_df, select, partition_by = compile_ast(nd.child)

    if isinstance(nd, verbs.Mutate | verbs.Summarize | verbs.Rename):
        if isinstance(nd, verbs.Rename):
            selected = set(select)
            names_to_consider = {
                col for uuid, col in name_in_df.items() if uuid not in selected
            }  # take it from name_in_df
            new_names = set(nd.name_map.values())
        else:
            nd_names_set = set(nd.names)
            # we need to do this before name_in_df is changed
            select = [
                uid
                for uid in (select if isinstance(nd, verbs.Mutate) else partition_by)
                if name_in_df[uid] not in nd_names_set
            ] + nd.uuids

            names_to_consider = (
                set(name_in_df[uid] for uid in partition_by)
                if isinstance(nd, verbs.Summarize)
                else set(name_in_df.values())
            )
            new_names = set(name for name in nd.names if name in names_to_consider)

        df, name_in_df = rename_overwritten_cols(
            new_names,
            df,
            name_in_df,
            names_to_consider=names_to_consider,
        )

    if isinstance(nd, verbs.Select):
        select = [col._uuid for col in nd.select]

    elif isinstance(nd, verbs.Rename):
        df = df.rename(nd.name_map)
        name_in_df = {uid: (nd.name_map[name] if name in nd.name_map else name) for uid, name in name_in_df.items()}

    elif isinstance(nd, verbs.Mutate):
        df = df.with_columns(
            **{name: compile_col_expr(value, name_in_df) for name, value in zip(nd.names, nd.values, strict=True)}
        )

        name_in_df.update({uid: name for uid, name in zip(nd.uuids, nd.names, strict=True)})

    elif isinstance(nd, verbs.Filter):
        if nd.predicates:
            df = df.filter(compile_col_expr(fil, name_in_df) for fil in nd.predicates)

    elif isinstance(nd, verbs.Arrange):
        order_by, descending, nulls_last = zip(
            *[compile_order(order, name_in_df) for order in nd.order_by], strict=True
        )
        df = df.sort(
            order_by,
            descending=descending,
            nulls_last=[False if nl is None else nl for nl in nulls_last],
            maintain_order=True,
        )

    elif isinstance(nd, verbs.Summarize):
        # We support usage of aggregated columns in expressions in summarize, but polars
        # creates arrays when doing that. Thus we unwrap the arrays when necessary.
        def has_path_to_leaf_without_agg(expr: ColExpr):
            if isinstance(expr, Col):
                return True
            if isinstance(expr, ColFn) and expr.op.ftype == Ftype.AGGREGATE:
                return False
            return any(has_path_to_leaf_without_agg(child) for child in expr.iter_children())

        aggregations = {}
        for name, val in zip(nd.names, nd.values, strict=True):
            # For some aggregations, a different polars function must be used if there
            # is not grouping. (In this case, `df.select` is called.)
            compiled = compile_col_expr(val, name_in_df, op_kwargs={"_empty_group_by": len(partition_by) == 0})
            if has_path_to_leaf_without_agg(val):
                compiled = compiled.first()
            aggregations[name] = compiled

        group_by = [name_in_df[uid] for uid in partition_by]
        # polars complains about an empty group_by
        if group_by:
            df = df.group_by(*group_by).agg(**aggregations)
        else:
            df = df.select(**aggregations)

        # we have to remove the columns here for the join hidden column rename to work
        # correctly (otherwise it would try to rename hidden columns that do not exist)
        name_in_df = {name: uuid for name, uuid in name_in_df.items() if name in partition_by}
        name_in_df.update({uid: name for name, uid in zip(nd.names, nd.uuids, strict=True)})
        partition_by = []

    elif isinstance(nd, verbs.SliceHead):
        df = df.slice(nd.offset, nd.n)

    elif isinstance(nd, verbs.GroupBy):
        new_group_by = [col._uuid for col in nd.group_by]
        partition_by = partition_by + new_group_by if nd.add else new_group_by

    elif isinstance(nd, verbs.Ungroup):
        partition_by = []

    elif isinstance(nd, verbs.Join):
        assert len(partition_by) == 0

        right_df, right_name_in_df, right_select, _ = compile_ast(nd.right)
        predicates = split_join_cond(nd.on)

        # check for name collisions among hidden columns
        # we maintain sensible names in the dataframe for all visible columns and try
        # to do so for hidden columns, too. If a hidden column has the same name as a
        # visible column, it gets a hash suffix. If two hidden columns collide, the
        # right one gets a hash.

        # visible columns
        right_df, right_name_in_df = rename_overwritten_cols(
            set(name_in_df[uid] for uid in select), right_df, right_name_in_df
        )
        df, name_in_df = rename_overwritten_cols(set(right_name_in_df[uid] for uid in right_select), df, name_in_df)

        # hidden columns
        right_df, right_name_in_df = rename_overwritten_cols(
            set(name_in_df.values()),
            right_df,
            right_name_in_df,
        )

        assert not set(right_name_in_df.keys()) & set(name_in_df.keys())
        name_in_df.update(right_name_in_df)

        eq_predicates = [pred for pred in predicates if pred.op == ops.equal]
        left_on, right_on = get_left_right_on(eq_predicates, name_in_df, right_name_in_df)

        # If there are only equality predicates, use normal join. Else use join_where
        if len(eq_predicates) == len(predicates):
            if len(predicates) == 0:
                # cross join, polars does not like an empty join condition
                df = df.join(right_df, how="cross")

            else:
                df = df.join(
                    right_df,
                    left_on=[compile_col_expr(col, name_in_df) for col in left_on],
                    right_on=[compile_col_expr(col, name_in_df) for col in right_on],
                    how=nd.how,
                    validate=nd.validate,
                    coalesce=False,
                )

        else:
            assert nd.how != "full"
            if nd.how == "left":
                df = df.with_columns(__INDEX__=pl.int_range(0, pl.len(), dtype=pl.Int64))

            joined = df.join_where(
                right_df,
                *(compile_col_expr(pred, name_in_df) for pred in predicates),
            ).with_columns(
                # polars deletes the right column in equality predicates...
                pl.col(name_in_df[left_col._uuid]).alias(name_in_df[right_col._uuid])
                for left_col, right_col in zip(left_on, right_on, strict=True)
                if isinstance(left_col, Col) and isinstance(right_col, Col)
            )

            if nd.how == "left":
                joined = df.join(joined, on="__INDEX__", how="left").drop("__INDEX__")

            df = joined

        select += right_select

    elif isinstance(nd, verbs.Union):
        assert len(partition_by) == 0

        right_df, right_name_in_df, right_select, _ = compile_ast(nd.right)

        # Drop all hidden columns from both dataframes before union
        # Note: The union verb already validates that visible column names match between tables,
        # so we can use left_col_names for both dataframes. This also ensures columns are in the
        # same order, which Polars requires for union operations.
        left_col_names = [name_in_df[uid] for uid in select]
        df = df.select(*left_col_names)
        right_df = right_df.select(*left_col_names)

        # Use pl.union if available (Polars >= 1.35), otherwise use pl.concat
        # pl.union is faster than pl.concat for union operations
        # distinct=True means UNION (remove duplicates), distinct=False means UNION ALL (keep duplicates)
        try:
            # Try to use pl.union (available in Polars >= 1.35)
            # pl.union takes a list of DataFrames/LazyFrames and has a distinct parameter
            if nd.distinct:
                # For UNION (distinct), use union with distinct=True
                df = pl.union([df, right_df], distinct=True)
            else:
                # For UNION ALL (not distinct), use union without distinct
                df = pl.union([df, right_df])
        except (AttributeError, TypeError):
            # Fall back to pl.concat for older Polars versions (< 1.35)
            if nd.distinct:
                # For UNION (distinct), we need to deduplicate
                # Polars doesn't have a direct UNION without ALL, so we concat and then distinct
                df = pl.concat([df, right_df]).unique()
            else:
                # For UNION ALL (not distinct), just concat
                df = pl.concat([df, right_df])

        # name_in_df and select remain the same (from left table)

    elif isinstance(nd, PolarsImpl):
        df = nd.df
        name_in_df = {col._uuid: col.name for col in nd.cols.values()}
        select = list(name_in_df.keys())
        partition_by = []

    return df, name_in_df, select, partition_by


with PolarsImpl.impl_store.impl_manager as impl:

    @impl(ops.mean)
    def _mean(x):
        return x.mean()

    @impl(ops.min)
    def _min(x):
        return x.min()

    @impl(ops.max)
    def _max(x):
        return x.max()

    @impl(ops.sum)
    def _sum(x):
        return x.sum()

    @impl(ops.all)
    def _all(x):
        return x.all()

    @impl(ops.any)
    def _any(x):
        return x.any()

    @impl(ops.is_null)
    def _is_null(x):
        return x.is_null()

    @impl(ops.is_not_null)
    def _is_not_null(x):
        return x.is_not_null()

    @impl(ops.fill_null)
    def _fill_null(x, y):
        return x.fill_null(y)

    @impl(ops.dt_year)
    def _dt_year(x):
        return x.dt.year().cast(pl.Int64)

    @impl(ops.dt_month)
    def _dt_month(x):
        return x.dt.month().cast(pl.Int64)

    @impl(ops.dt_day)
    def _dt_day(x):
        return x.dt.day().cast(pl.Int64)

    @impl(ops.dt_hour)
    def _dt_hour(x):
        return x.dt.hour().cast(pl.Int64)

    @impl(ops.dt_minute)
    def _dt_minute(x):
        return x.dt.minute().cast(pl.Int64)

    @impl(ops.dt_second)
    def _dt_second(x):
        return x.dt.second().cast(pl.Int64)

    @impl(ops.dt_millisecond)
    def _dt_millisecond(x):
        return x.dt.millisecond().cast(pl.Int64)

    @impl(ops.dt_microsecond)
    def _dt_microsecond(x):
        return x.dt.microsecond().cast(pl.Int64)

    @impl(ops.dt_day_of_week)
    def _dt_day_of_week(x):
        return x.dt.weekday().cast(pl.Int64)

    @impl(ops.dt_day_of_year)
    def _dt_day_of_year(x):
        return x.dt.ordinal_day().cast(pl.Int64)

    @impl(ops.dur_days)
    def _dur_days(x):
        return x.dt.total_days()

    @impl(ops.dur_hours)
    def _dur_hours(x):
        return x.dt.total_hours()

    @impl(ops.dur_minutes)
    def _dur_minutes(x):
        return x.dt.total_minutes()

    @impl(ops.dur_seconds)
    def _dur_seconds(x):
        return x.dt.total_seconds()

    @impl(ops.dur_milliseconds)
    def _dur_milliseconds(x):
        return x.dt.total_milliseconds()

    @impl(ops.row_number)
    def _row_number():
        return pl.int_range(start=1, end=pl.len() + 1, dtype=pl.Int64)

    @impl(ops.rank)
    def _rank(x):
        return x.rank("min").cast(pl.Int64)

    @impl(ops.dense_rank)
    def _dense_rank(x):
        return x.rank("dense").cast(pl.Int64)

    @impl(ops.shift)
    def _shift(x, n, fill_value=None):
        return x.shift(n, fill_value=fill_value)

    @impl(ops.is_in)
    def _is_in(x, *values):
        if len(values) == 0:
            return pl.lit(False)
        return pl.any_horizontal(x == val for val in values)

    @impl(ops.str_contains)
    def _str_contains(x, y, allow_regex, true_if_regex_unsupported):
        _ = true_if_regex_unsupported
        return x.str.contains(y, literal=not pl.select(allow_regex).item())

    @impl(ops.str_starts_with)
    def _str_starts_with(x, y):
        return x.str.starts_with(y)

    @impl(ops.str_ends_with)
    def _str_ends_with(x, y):
        return x.str.ends_with(y)

    @impl(ops.str_lower)
    def _str_lower(x):
        return x.str.to_lowercase()

    @impl(ops.str_upper)
    def _str_upper(x):
        return x.str.to_uppercase()

    @impl(ops.str_replace_all)
    def _str_replace_all(x, to_replace, replacement):
        return x.str.replace_all(to_replace, replacement)

    @impl(ops.str_len)
    def _str_len(x):
        return x.str.len_chars().cast(pl.Int64)

    @impl(ops.str_strip)
    def _str_strip(x):
        return x.str.strip_chars()

    @impl(ops.str_slice)
    def _str_slice(x, offset, length):
        return x.str.slice(offset, length)

    @impl(ops.count)
    def _count(x):
        return x.count().cast(pl.Int64)

    @impl(ops.count_star)
    def _len():
        return pl.len().cast(pl.Int64)

    @impl(ops.horizontal_max)
    def _horizontal_max(*x):
        return pl.max_horizontal(*x)

    @impl(ops.horizontal_min)
    def _horizontal_min(*x):
        return pl.min_horizontal(*x)

    @impl(ops.round)
    def _round(x, digits):
        digits = pl.select(digits).item()
        if digits < 0:
            return (x / (10**-digits)).round() * (10**-digits)
        return x.round(digits)

    @impl(ops.exp)
    def _exp(x):
        return x.exp()

    @impl(ops.log)
    def _log(x):
        return x.log()

    @impl(ops.floor)
    def _floor(x):
        return x.floor()

    @impl(ops.ceil)
    def _ceil(x):
        return x.ceil()

    @impl(ops.str_to_datetime)
    def _str_to_datetime(x):
        return x.str.to_datetime()

    @impl(ops.str_to_date)
    def _str_to_date(x):
        return x.str.to_date()

    @impl(ops.floordiv)
    def _floordiv(lhs, rhs):
        result_sign = (lhs < 0) ^ (rhs < 0)
        return (abs(lhs) // abs(rhs)) * pl.when(result_sign).then(-1).otherwise(1)
        # TODO: test some alternatives if this is too slow

    @impl(ops.mod)
    def _mod(lhs, rhs):
        return lhs % (abs(rhs) * pl.when(lhs >= 0).then(1).otherwise(-1))
        # TODO: see whether the following is faster:
        # pl.when(lhs >= 0).then(lhs % abs(rhs)).otherwise(lhs % -abs(rhs))

    @impl(ops.is_inf)
    def _is_inf(x):
        return x.is_infinite()

    @impl(ops.is_not_inf)
    def _is_not_inf(x):
        return x.is_finite()

    @impl(ops.is_nan)
    def _is_nan(x):
        return x.is_nan()

    @impl(ops.is_not_nan)
    def _is_not_nan(x):
        return x.is_not_nan()

    @impl(ops.coalesce)
    def _coalesce(*x):
        return pl.coalesce(*x)

    @impl(ops.pow, Int(), Int())
    def _pow(x, y):
        return x.cast(pl.Float64()) ** y

    @impl(ops.str_join)
    def _str_join(x, delim):
        return x.str.join(pl.select(delim).item())

    @impl(ops.cum_sum)
    def _cum_sum(x: pl.Expr):
        return x.cum_sum().fill_null(strategy="forward")

    @impl(ops.list_agg)
    def _list_agg(x, *, _empty_group_by: bool):
        if _empty_group_by:
            return x.implode()
        return x

    @impl(ops.sin)
    def _sin(x):
        return x.sin()

    @impl(ops.cos)
    def _cos(x):
        return x.cos()

    @impl(ops.tan)
    def _tan(x):
        return x.tan()

    @impl(ops.asin)
    def _asin(x):
        return x.arcsin()

    @impl(ops.acos)
    def _acos(x):
        return x.arccos()

    @impl(ops.atan)
    def _atan(x):
        return x.arctan()

    @impl(ops.sqrt)
    def _sqrt(x):
        return x.sqrt()

    @impl(ops.cbrt)
    def _cbrt(x):
        return x.cbrt()

    @impl(ops.log10)
    def _log10(x):
        return x.log10()

    @impl(ops.clip)
    def _clip(x, lower, upper):
        return x.clip(lower, upper)

    @impl(ops.rand)
    def _rand():
        return pl.int_range(pl.len()).map_elements(lambda x: random.random(), pl.Float64())



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from datetime import date, datetime

import polars as pl
import pytest

from tests.fixtures.backend import BACKEND_MARKS, flatten
from tests.util.backend import BACKEND_TABLES

dataframes = {
    "df1": pl.DataFrame(
        {
            "col1": [1, 2, 3, 4],
            "col2": ["a", "baa", "c", "d"],
            "cnull": [None, 2, None, None],
        }
    ),
    "df2": pl.DataFrame(
        {
            "col1": [1, 2, -2, 4, 5, 6],
            "col2": [2, 2, 0, 0, 2, None],
            "col3": [0.0, -0.1, 0.2, 0.3, 0.01, 0.02],
        }
    ),
    "df3": pl.DataFrame(
        {  # tests rely on col4 having a unique ordering
            "col1": [0, 0, -1, -1000, 1, 1, 0, 1, 4, 2, 2, 2],
            "col2": [0, 0, 10, 1, 0, 5, 1, 1, 0, 0, 1, 1],
            "col3": [0, 1, 2, 3, 0, 1, 2, 3, 1, 1, 2, 3],
            "col4": [-1729, 2, 1, 3, 4, 5, 13, 7, 8, 9, 10, 11],
            "col5": list("abcdafghijkk"),
            "col6": [
                "a" * 30000,
            ]
            * 12,
            "col7": [
                "b" * 70000,
            ]
            * 12,
        }
    ),
    "df4": pl.DataFrame(
        {  # tests rely on col4 having a unique ordering
            "col1": [None, 0, 0, 0, 0, None, 1, 1, 1, 2, 3, 2, 2],
            "col2": [0, 20, 1, 1, 0, 0, 1, None, 1, 0, 0, 0, 1],
            "col3": [None, None, None, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
            "col4": [None, -42, 2, 33, 4, 5, -10, 7, 8, 90, 10, 11, 12],
            "col5": list("abcdefghijkl") + [None],
            "col6": [
                "a" * 30000,
            ]
            * 12
            + [None],
            "col7": [
                "b" * 70000,
            ]
            * 12
            + [None],
        }
    ),
    "df_strings": pl.DataFrame(
        {
            "col1": [
                "",
                " ",
                "xyzzzz123",
                " x ",
                "foo",
                "FooBarfooofoo",
                "abracadabra",
                None,
                "--+011x",
                None,
                "_ %",
                "            ",
            ],
            "col2": [
                "",
                "test_%",
                "xyz",
                " x",
                "FOO33",
                "FooBar",
                "AbracadabA",
                "",
                "barAb",
                None,
                "% _.AbAbAb",
                "       ",
            ],
            "c": [
                "4352.0",
                "-21",
                "-1.121222",
                "3.313",
                None,
                "-444",
                "5.3",
                "1.33333",
                "-0.000",
                "-0.0",
                "0.0",
                "  8   ",
            ],
            "d": [
                None,
                "-123124",
                "21241",
                "010101",
                "0",
                "1",
                "-12",
                "42",
                "197",
                "1729",
                "-100110",
                "   -56 ",
            ],
            "e": [
                "abbabbabbabba",
                "------",
                "012",
                "",
                "",
                None,
                None,
                " AbAbA..",
                "-354.2",
                "??",
                "11",
                "$&/)",
            ],
            "gb": ["a", "b", "a", "a", "0", "", "", "a", "", "c", "d", "b"],
            "long": [
                "a" * 30000,
            ]
            * 12,
            "clob": [
                "b" * 70000,
            ]
            * 12,
        }
    ),
    "df_datetime": pl.DataFrame(
        {
            "col1": [
                datetime(2000, 1, 1),
                datetime(2000, 1, 1, 12, 15, 59),
                datetime(1970, 1, 1, 12, 30, 30, 987_123),
                datetime(1700, 4, 27, 1, 2, 3),
                None,
                datetime(2250, 12, 24, 23, 0, 0),
                datetime(2023, 7, 30, 10, 16, 13),
                datetime(2004, 12, 31),
                datetime(1970, 1, 5),
            ],
            "col2": [
                datetime(2023, 7, 31, 10, 16, 13),
                datetime(2023, 7, 31, 10, 16, 14),
                datetime(1900, 1, 2, 3, 4, 5),
                datetime(1988, 5, 3, 12, 4, 54),
                None,
                datetime(2023, 7, 31, 10, 16, 13),
                None,
                datetime(2004, 12, 31, 23, 59, 59, 456_789),
                datetime(1970, 1, 1),
            ],
            "cdate": [
                date(2017, 3, 2),
                date(1998, 1, 12),
                date(1999, 12, 31),
                date(2024, 9, 23),
                date(2018, 8, 13),
                None,
                date(2010, 5, 1),
                date(2016, 2, 27),
                date(2000, 1, 1),
            ],
            # "cdur": [
            #     None,
            #     timedelta(1, 4, 2, 5),
            #     timedelta(0, 11, 14, 10000),
            #     timedelta(12, 2, 3),
            #     timedelta(4, 3, 1, 2, 3, 4),
            #     timedelta(0, 0, 0, 0, 1),
            #     timedelta(0, 1, 0, 1, 0, 1),
            #     None,
            #     timedelta(),
            # ],
        }
    ),
    "df_num": pl.DataFrame(
        {
            "a": [0.4, -1.1, -0.0, 0.0, 9.0, 2.0, -344.0053, -1000.0],
            "b": [None, 2.0, 0.0, -11.0, 4.0, 19.0, -5190.0, 2000000.0],
            "c": [0.0, None, None, 2.9, -0.0, 10.0, -10.0, 3.1415926535],
            "d": [None, 2352.0230, 0.577, 901234, -6.0, 4.0, None, -99.0],
            "e": [1.0, 2.0, 3.0, 4.99, -442.0, 6.0, 7.0, 500.0],
            "f": [3.0, None, 0.0, 4.3, 10.0, -1.2, -9999.1, -34.1],
            "g": [-5.5, None, None, 1.100212, -3.412351, 1000.4252, 0.0, -1.6],
            "zero": [0.0, -0.0] * 4,
            "pos": [
                1.123,
                1297.324,
                7.5,
                1e50 + 54356346912.332131,
                912.097,
                1e-51,
                0.12,
                5002352.434,
            ],
            "neg": [
                -9623.1,
                -0.1,
                -1.0,
                -1e19 - 394734729923737552.5,
                -5.5,
                -0.0001,
                -1.2e-39,
                -6699917733.1242,
            ],
            "null_s": [0.0, None, None, None, None, None, None, None],
        }
    ),
    "df_int": pl.DataFrame(
        {
            "a": [3, 1, 0, -12, 4, 5, 1 << 20, 5],
            "b": [-23, 18282, -42, 1729, None, -2323, 11, 1],
            "pos": [1 << 31, (1 << 23) - 1, 2, 129879000, 233, 9223222, 1, 5],
            "neg": [
                -(1 << 22),
                -(1 << 31),
                -(1 << 26) + 1,
                -1,
                -234,
                -3,
                -22,
                -23938333,
            ],
            "null_s": [0] + [None] * 7,
        }
    ),
    "df_bool": pl.DataFrame(
        {
            "a": [True, True, True, False, False, None],
            "b": [True, False, None, False, None, None],
        },
    ),
}

# compare one dataframe and one SQL backend to all others
# (some tests are ignored if either backend does not support a feature)
reference_backends = ["polars", "duckdb"]


def pytest_generate_tests(metafunc: pytest.Metafunc):
    pl.Config.set_tbl_cols(20)
    pl.Config.set_tbl_rows(40)
    pl.Config.set_tbl_width_chars(160)
    # Parametrize tests based on `backends` and `skip_backends` mark.

    backends = dict.fromkeys(BACKEND_TABLES)
    if mark := metafunc.definition.get_closest_marker("backends"):
        backends = dict.fromkeys(mark.args)
    if mark := metafunc.definition.get_closest_marker("skip_backends"):
        for backend in mark.args:
            if backend in backends:
                del backends[backend]

    backends = {k: i for i, k in enumerate(backends)}

    backend_combinations = [
        (reference_backend, backend) for reference_backend in reference_backends for backend in backends
    ]

    params = []
    table_names = [name for name in metafunc.fixturenames if name in dataframes]
    for reference_backend, backend in backend_combinations:
        # Skip some redundant backend combinations
        if reference_backend == backend:
            continue
        if reference_backend not in backends:
            continue
        if backend in reference_backends:
            if backends[backend] < backends[reference_backend]:
                continue

        marks = list(
            flatten(
                [
                    BACKEND_MARKS.get(reference_backend, ()),
                    BACKEND_MARKS.get(backend, ()),
                ]
            )
        )

        arguments = [(reference_backend, backend)] * len(table_names)

        params.append(
            pytest.param(
                *arguments,
                marks=marks,
                id=f"{reference_backend}-{backend}",
            )
        )

    param_names = ",".join([f"{name}" for name in table_names])
    return metafunc.parametrize(param_names, params, indirect=True)


def generate_df_fixtures():
    def gen_fixture(table_name):
        @pytest.fixture(scope="function")
        def table_fixture(request):
            df = dataframes[table_name]
            name = table_name

            table_x = BACKEND_TABLES[request.param[0]](df, name)
            table_y = BACKEND_TABLES[request.param[1]](df, name)
            return table_x, table_y

        return table_fixture

    for name in dataframes:
        globals()[f"{name}"] = gen_fixture(name)


generate_df_fixtures()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import itertools
from collections.abc import Callable, Iterable
from uuid import UUID

from pydiverse.transform._internal.util.warnings import warn


class AstNode:
    __slots__ = ["name"]

    name: str

    def clone(self) -> "AstNode":
        return self._clone()[0]

    def _clone(
        self,
    ) -> tuple["AstNode", dict["AstNode", "AstNode"], dict[UUID, UUID]]: ...

    def iter_subtree_postorder(self) -> Iterable["AstNode"]: ...

    def iter_subtree_preorder(self) -> Iterable["AstNode"]: ...

    def __repr__(self) -> str:
        return self.short_name()

    # Formatted, almost source code like representation of the AST.
    def ast_repr(self, verb_depth: int = -1, expr_depth: int = -1) -> str:
        from pydiverse.transform._internal.backend.table_impl import TableImpl
        from pydiverse.transform._internal.tree.col_expr import Col
        from pydiverse.transform._internal.tree.verbs import Alias, Verb

        def next_nd(root: AstNode, cond: Callable[["AstNode"], bool]):
            for nd in root.iter_subtree_preorder():
                if cond(nd):
                    return nd

        source_tbls = set(nd for nd in self.iter_subtree_preorder() if isinstance(nd, TableImpl | Alias))
        # Add required ASTs from aligned columns (they need not be in the subtree of
        # `self`)
        source_tbls |= set(
            next_nd(col._ast, lambda x: isinstance(x, Alias | TableImpl))
            for col in itertools.chain(
                *(nd.iter_col_nodes() for nd in self.iter_subtree_preorder() if isinstance(nd, Verb))
            )
            if isinstance(col, Col)
        )

        table_display_name_map: dict[TableImpl, str] = dict()
        used = set()
        for nd in source_tbls:
            display_name = nd.name or "tbl"
            # try to achieve valid python identifier names
            display_name = display_name.replace(" ", "_").replace(".", "_").replace("-", "_")

            if display_name not in used:
                used.add(display_name)
                table_display_name_map[nd] = display_name
            else:
                cnt = 1
                while f"{display_name}_{cnt}" in used:
                    cnt += 1
                table_display_name_map[nd] = f"{display_name}_{cnt}"
                used.add(f"{display_name}_{cnt}")

        # Find the last source table / alias for every node in the AST and use the
        # corresponding name.
        for nd in self.iter_subtree_postorder():
            table_display_name_map[nd] = table_display_name_map[next_nd(nd, lambda x: x in table_display_name_map)]

            if isinstance(nd, Verb):
                for col in nd.iter_col_nodes():
                    if isinstance(col, Col):
                        table_display_name_map[col._ast] = table_display_name_map[
                            next_nd(col._ast, lambda x: x in table_display_name_map)
                        ]

        unformatted = "\n".join(
            f"{display_name} = {tbl._table_def_repr()}"
            for tbl, display_name in table_display_name_map.items()
            if isinstance(tbl, TableImpl)
        ) + ("\n\n(" + self._unformatted_ast_repr(verb_depth, expr_depth, table_display_name_map) + ")")
        try:
            import black

            formatted = black.format_str(unformatted, mode=black.Mode(line_length=120))
            return formatted
        except Exception:
            warn("Could not format AST representation with `black`.")
            return unformatted

    def short_name(self) -> str:
        raise NotImplementedError()

    # Recursive, builds up the verb chain.
    def _unformatted_ast_repr(self, verb_depth: int, expr_depth: int, display_name_map):
        raise NotImplementedError()

    # Just the verb call of a single verb, without `>>`.
    def _ast_node_repr(self, expr_depth: int, display_name_map):
        raise NotImplementedError()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import Bool

bool_and = Operator(
    "__and__",
    Signature(Bool(), Bool(), return_type=Bool()),
    doc="""
Boolean AND (__and__)

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [True, True, True, False, False, None],
...         "b": [True, False, None, False, None, None],
...     },
...     name="bool table",
... )
>>> t >> mutate(x=t.a & t.b) >> show()
Table bool table, backend: PolarsImpl
shape: (6, 3)
┌───────┬───────┬───────┐
│ a     ┆ b     ┆ x     │
│ ---   ┆ ---   ┆ ---   │
│ bool  ┆ bool  ┆ bool  │
╞═══════╪═══════╪═══════╡
│ true  ┆ true  ┆ true  │
│ true  ┆ false ┆ false │
│ true  ┆ null  ┆ null  │
│ false ┆ false ┆ false │
│ false ┆ null  ┆ false │
│ null  ┆ null  ┆ null  │
└───────┴───────┴───────┘
""",
)

bool_or = Operator(
    "__or__",
    Signature(Bool(), Bool(), return_type=Bool()),
    doc="""
Boolean OR (__or__)

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [True, True, True, False, False, None],
...         "b": [True, False, None, False, None, None],
...     },
...     name="bool table",
... )
>>> t >> mutate(x=t.a | t.b) >> show()
Table bool table, backend: PolarsImpl
shape: (6, 3)
┌───────┬───────┬───────┐
│ a     ┆ b     ┆ x     │
│ ---   ┆ ---   ┆ ---   │
│ bool  ┆ bool  ┆ bool  │
╞═══════╪═══════╪═══════╡
│ true  ┆ true  ┆ true  │
│ true  ┆ false ┆ true  │
│ true  ┆ null  ┆ true  │
│ false ┆ false ┆ false │
│ false ┆ null  ┆ null  │
│ null  ┆ null  ┆ null  │
└───────┴───────┴───────┘
""",
)

bool_xor = Operator(
    "__xor__",
    Signature(Bool(), Bool(), return_type=Bool()),
    doc="""
Boolean XOR (__xor__)

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [True, True, True, False, False, None],
...         "b": [True, False, None, False, None, None],
...     },
...     name="bool table",
... )
>>> t >> mutate(x=t.a ^ t.b) >> show()
Table bool table, backend: PolarsImpl
shape: (6, 3)
┌───────┬───────┬───────┐
│ a     ┆ b     ┆ x     │
│ ---   ┆ ---   ┆ ---   │
│ bool  ┆ bool  ┆ bool  │
╞═══════╪═══════╪═══════╡
│ true  ┆ true  ┆ false │
│ true  ┆ false ┆ true  │
│ true  ┆ null  ┆ null  │
│ false ┆ false ┆ false │
│ false ┆ null  ┆ null  │
│ null  ┆ null  ┆ null  │
└───────┴───────┴───────┘
""",
)

bool_invert = Operator(
    "__invert__",
    Signature(Bool(), return_type=Bool()),
    doc="""
Boolean inversion (__invert__)

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [True, True, True, False, False, None],
...         "b": [True, False, None, False, None, None],
...     },
...     name="bool table",
... )
>>> t >> mutate(x=~t.a) >> show()
Table bool table, backend: PolarsImpl
shape: (6, 3)
┌───────┬───────┬───────┐
│ a     ┆ b     ┆ x     │
│ ---   ┆ ---   ┆ ---   │
│ bool  ┆ bool  ┆ bool  │
╞═══════╪═══════╪═══════╡
│ true  ┆ true  ┆ false │
│ true  ┆ false ┆ false │
│ true  ┆ null  ┆ false │
│ false ┆ false ┆ true  │
│ false ┆ null  ┆ true  │
│ null  ┆ null  ┆ null  │
└───────┴───────┴───────┘
""",
)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from itertools import chain

import pytest

# Pytest markers associated with specific backend
BACKEND_MARKS = {
    "postgres": pytest.mark.postgres,
    "mssql": pytest.mark.mssql,
    "ibm_db2": pytest.mark.ibm_db2,
}


def with_backends(*backends):
    """Decorator to run a test with a specific set of backends

    :param backends: Names of the backends to use.
    """
    return pytest.mark.backends(*flatten(backends))


def skip_backends(*backends):
    """Decorator to skip running a test with a specific set of backends"""
    return pytest.mark.skip_backends(*flatten(backends))


def flatten(it):
    """Flatten an iterable"""
    if isinstance(it, list | tuple):
        yield from chain(*map(flatten, it))
    else:
        yield it



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import functools
import os
import tempfile
from pathlib import Path

import polars as pl
import sqlalchemy as sqa

from pydiverse.transform._internal.backend.targets import SqlAlchemy
from pydiverse.transform._internal.pipe.table import Table


def _cached_table(fn):
    cache = {}

    @functools.wraps(fn)
    def wrapped(df: pl.DataFrame, name: str):
        if name in cache:
            return cache[name]

        impl = fn(df, name)
        cache[name] = impl

        return impl

    return wrapped


@_cached_table
def polars_table(df: pl.DataFrame, name: str):
    return Table(df, name=name)


_sql_engine_cache = {}


def sql_table(
    df: pl.DataFrame,
    name: str,
    url: str,
    dtypes_map: dict | None = None,
    sql_dtypes: dict | None = None,
    fix_sql_dtypes: dict | None = None,
    dialect_infix="",
):
    import sqlalchemy as sqa

    global _sql_engine_cache

    dtypes_map = dtypes_map or {}
    dtypes_map[pl.Float64] = sqa.Double

    if url in _sql_engine_cache:
        engine = _sql_engine_cache[url]
    else:
        engine = sqa.create_engine(url)
        _sql_engine_cache[url] = engine

    sql_dtypes = sql_dtypes or {}
    for col, dtype in zip(df.columns, df.dtypes, strict=True):
        if dtype in dtypes_map and col not in sql_dtypes:
            sql_dtypes[col] = dtypes_map[dtype]

    df.write_database(name, engine, if_table_exists="replace", engine_options={"dtype": sql_dtypes})
    if fix_sql_dtypes is not None and len(fix_sql_dtypes) > 0:
        # this is a hack to fix sql types after creation of the table
        # the main reason for this is that ibm_db_sa renders sqa.boolean as SMALLINT
        # (https://github.com/ibmdb/python-ibmdbsa/issues/161)
        with engine.connect() as conn:
            for col, dtype in fix_sql_dtypes.items():
                conn.execute(sqa.text(f"ALTER TABLE {name} ALTER COLUMN {col} {dialect_infix} {dtype}"))
            conn.execute(sqa.text(f"call sysproc.admin_cmd('REORG TABLE {name}')"))
            conn.commit()
    return Table(name, SqlAlchemy(engine))


@_cached_table
def sqlite_table(df: pl.DataFrame, name: str):
    return sql_table(df, name, "sqlite:///:memory:")


@_cached_table
def duckdb_table(df: pl.DataFrame, name: str):
    return sql_table(df, name, "duckdb:///:memory:")


@_cached_table
def duckdb_parquet_table(df: pl.DataFrame, name: str):
    import sqlalchemy as sqa

    global _sql_engine_cache

    if "duckdb_parquet" in _sql_engine_cache:
        engine = _sql_engine_cache["duckdb_parquet"]
    else:
        engine = sqa.create_engine("duckdb:///:memory:")
    _sql_engine_cache["duckdb_parquet"] = engine
    path = Path(tempfile.gettempdir()) / "pdtransform" / "tests"
    file = path / f"{name}.parquet"
    os.makedirs(path, exist_ok=True)
    if file.exists():
        os.unlink(file)
    df.write_parquet(file)

    with engine.connect() as conn:
        conn.execute(sqa.text(f"DROP VIEW IF EXISTS {name}"))
        conn.execute(sqa.text(f"CREATE VIEW {name} AS SELECT * FROM read_parquet('{path / f'{name}.parquet'}')"))
        conn.commit()

    return Table(name, SqlAlchemy(engine))


@_cached_table
def postgres_table(df: pl.DataFrame, name: str):
    url = "postgresql://sa:Pydiverse23@127.0.0.1:6543"

    return sql_table(df, name, url)


@_cached_table
def mssql_table(df: pl.DataFrame, name: str):
    from sqlalchemy.dialects.mssql import DATETIME2

    url = "mssql+pyodbc://sa:PydiQuant27@127.0.0.1:1433/master?driver=ODBC+Driver+18+for+SQL+Server&encrypt=no"
    return sql_table(df, name, url, dtypes_map={pl.Datetime(): DATETIME2()})


@_cached_table
def ibm_db2_table(df: pl.DataFrame, name: str):
    url = "db2+ibm_db://db2inst1:password@localhost:50000/testdb"

    map = {}
    fix_map = {}
    for col, dtype in zip(df.columns, df.dtypes, strict=True):
        if dtype == pl.String:
            max_length = df[col].str.len_chars().max()
            if max_length > 32_672:
                map[col] = sqa.CLOB()
            elif max_length > 256:
                map[col] = sqa.VARCHAR(32_672)
            else:
                map[col] = sqa.VARCHAR(256)
        if dtype == pl.Boolean:
            fix_map[col] = sqa.Boolean()

    return sql_table(
        df,
        name,
        url,
        sql_dtypes=map,
        fix_sql_dtypes=fix_map,
        dialect_infix="SET DATA TYPE",
    )


BACKEND_TABLES = {
    "polars": polars_table,
    "duckdb": duckdb_table,
    "duckdb_parquet": duckdb_parquet_table,
    "sqlite": sqlite_table,
    "postgres": postgres_table,
    "mssql": mssql_table,
    "ibm_db2": ibm_db2_table,
}



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform.extended import *
from tests.fixtures.backend import skip_backends
from tests.util import assert_result_equal


@skip_backends("ibm_db2")  # IBM DB2 is not good with whitespaces
def test_eq_whitespace(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 == " "))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 == C.col2))


def test_eq(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 == "foo"))
    assert_result_equal(
        df_strings,
        lambda t: t >> filter(C.col1.str.replace_all(" ", "") == C.col2.str.replace_all(" ", "")),
    )


@skip_backends("ibm_db2")  # IBM DB2 is not good with whitespaces
def test_nq_whitespace(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 != " "))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 != C.col2))


def test_nq(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 != "foo"))
    assert_result_equal(
        df_strings,
        lambda t: t >> filter(C.col1.str.replace_all(" ", "") != C.col2.str.replace_all(" ", "")),
    )


def test_lt(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 < " x"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 < "E"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 < C.col2))


@skip_backends("ibm_db2")  # IBM DB2 is not good with whitespaces
def test_gt_whitespace(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 > " x"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 > C.col2))


def test_gt(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 > "E"))
    assert_result_equal(
        df_strings,
        lambda t: t >> filter(C.col1.str.replace_all(" ", "") > C.col2.str.replace_all(" ", "")),
    )


@skip_backends("ibm_db2")  # IBM DB2 is not good with whitespaces
def test_le_whitespace(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            col1_le_c=C.col1 <= C.c,
            col1_le_col2=t.col1 <= t.col2,
            d_le_c=t.d <= t.c,
        ),
    )
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 <= C.col2))


def test_le(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            col1_le_c=C.col1 <= C.c,
            col1_le_col2=t.col1.str.replace_all(" ", "") <= t.col2.str.replace_all(" ", ""),
            d_le_c=t.d <= t.c,
        ),
    )
    assert_result_equal(
        df_strings,
        lambda t: t >> filter(C.col1.str.replace_all(" ", "") <= C.col2.str.replace_all(" ", "")),
    )


def test_ge(df_strings):
    assert_result_equal(df_strings, lambda t: t >> mutate(col1_ge_col2=C.col1 >= C.col2))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 >= C.col2))


def test_strip(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.strip(),
            y=C.col2.str.strip(),
        ),
    )


def test_string_length(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.len(),
            y=C.col2.str.len(),
        ),
    )


def test_upper(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.upper(),
            y=C.col2.str.upper(),
        ),
    )


def test_lower(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.lower(),
            y=C.col2.str.lower(),
        ),
    )


def test_replace_all(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.replace_all(" ", "").str.replace_all("foo", "fOO"),
            y=C.col2.str.replace_all("Ab", "ab"),
            z=C.e.str.replace_all("abba", "#"),
            q=C.e.str.replace_all("--", "="),
        ),
    )


def test_starts_with(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.starts_with("foo") | C.col1.str.starts_with(" "),
            y=C.col2.str.starts_with("test") | C.col2.str.starts_with("Abra"),
            underscore=C.col1.str.starts_with("_"),
            percent=C.col2.str.starts_with("%"),
        ),
    )


def test_ends_with(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.ends_with("Bar") | C.col1.str.ends_with(" "),
            y=C.col2.str.ends_with("_%") | C.col2.str.ends_with("Bar"),
            percent=C.col1.str.ends_with("%"),
            underscore=C.col2.str.ends_with("_"),
        ),
    )


def test_contains(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.contains(" ") | C.col1.str.contains("Foo"),
            y=C.col2.str.contains("st_") | C.col2.str.contains("bar"),
            percent=C.col1.str.contains("%", allow_regex=False),
            underscore=C.col2.str.contains("_"),
        ),
    )


@skip_backends("sqlite", "mssql")
def test_contains_regex(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=t.col1.str.contains("."),
            y=(t.col2 + t.c).str.contains("[^abc]"),
            z=t.e.str.contains(""),
            h=t.e.str.contains("A.", allow_regex=False),
        ),
    )


@skip_backends("ibm_db2")  # IBM DB2 is not good with whitespaces
def test_slice_whitespace(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            u=t.col1.str.slice(1, 3),
            v=t.col2.str.slice(t.col1.str.len() % (t.col2.str.len() + 1), 42),
            w=t.col1.str.slice(2, t.col1.str.len()),
        ),
    )


def test_slice(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            u=t.col1.str.replace_all(" ", "").str.slice(1, 3).str.replace_all(" ", ""),
            v=t.col2.str.replace_all(" ", "")
            .str.slice(t.col1.str.len() % (t.col2.str.len() + 1), 42)
            .str.replace_all(" ", ""),
            w=t.col1.str.replace_all(" ", "").str.slice(2, t.col1.str.len()).str.replace_all(" ", ""),
        ),
    )


@skip_backends("ibm_db2")  # ibm_db_sa does not support aggregate_order_by
def test_str_join(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> group_by(t.e)
        >> summarize(con=t.c.str.join(", ", arrange=[t.d.nulls_first(), t.c.nulls_last()])),
    )

    assert_result_equal(
        df_strings,
        lambda t: t >> group_by(t.gb) >> summarize(y=t.col1.str.join(arrange=t.col2)),
    )


@skip_backends("mssql", "ibm_db2")  # IBM DB2 is not good with whitespaces
def test_str_arrange_whitespace(df_strings):
    def bind(col):
        assert_result_equal(
            df_strings,
            lambda t: t >> arrange(t[col].nulls_last(), t.c.nulls_last()),
            check_row_order=True,
        )

    for col in ["col1", "col2", "c", "d", "e", "gb"]:
        bind(col)


def test_str_arrange(df_strings):
    def bind(col):
        assert_result_equal(
            df_strings,
            lambda t: t >> arrange(t[col].str.replace_all(" ", "").nulls_last(), t.c.nulls_last()),
            check_row_order=True,
        )

    for col in ["col1", "col2", "c", "d", "e", "gb"]:
        bind(col)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import inspect
from functools import wraps

import pandas as pd
import polars as pl

from pydiverse.transform._internal import errors
from pydiverse.transform._internal.pipe.table import Table
from pydiverse.transform._internal.tree.col_expr import Col, ColExpr, EvalAligned


def aligned(fn=None, *, with_: str | None = None):
    """
    Decorator that automatically applies :doc:`pydiverse.transform.eval_aligned` to the
    return value of a function.

    :param with_:
        The table or column to align with.

    Examples
    --------
    >>> @aligned(with_="col")
    ... def reverse_col(col: pdt.Col) -> pdt.ColExpr:
    ...     return col.export(Polars).reverse()
    ...
    >>> t = pdt.Table(
    ...     {
    ...         "a": [1, 2, 3, 4],
    ...         "b": [2, 5, 16, 3],
    ...     },
    ...     name="t",
    ... )
    >>> t >> mutate(r=reverse_col(t.a)) >> show()
    Table `t` (backend: polars)
    shape: (4, 3)
    ┌─────┬─────┬─────┐
    │ a   ┆ b   ┆ r   │
    │ --- ┆ --- ┆ --- │
    │ i64 ┆ i64 ┆ i64 │
    ╞═════╪═════╪═════╡
    │ 1   ┆ 2   ┆ 4   │
    │ 2   ┆ 5   ┆ 3   │
    │ 3   ┆ 16  ┆ 2   │
    │ 4   ┆ 3   ┆ 1   │
    └─────┴─────┴─────┘
    """

    errors.check_arg_type(str | None, "aligned", "with_", with_)

    def decorator(fn):
        signature = inspect.signature(fn)
        if with_ is not None and with_ not in signature.parameters:
            raise ValueError(f"function `{fn.__name__}` has no argument named `{with_}`")

        @wraps(fn)
        def wrapper(*args, **kwargs):
            if with_ is not None:
                bound_sig = signature.bind(*args, **kwargs)
                bound_sig.apply_defaults()
                with_obj = bound_sig.arguments[with_]
            else:
                with_obj = None

            return eval_aligned(fn(*args, **kwargs), with_=with_obj)

        return wrapper

    if fn is not None:
        return decorator(fn)

    return decorator


def eval_aligned(val: ColExpr | pl.Series | pd.Series, with_: Table | Col | None = None) -> EvalAligned:
    """
    Allows to evaluate a column expression containing columns from different tables and
    to use polars / pandas Series in column expressions.

    :param val:
        The expression or polars / pandas Series to be aligned.

    :param with_:
        The table or column to align with.

    Examples
    --------
    Usage of a polars `Series` in a column expressions (the same works for pandas
    `Series`):

    >>> import polars as pl
    >>> t = pdt.Table(
    ...     {
    ...         "a": [1, 2, 3, 4],
    ...         "b": [2, 5, 16, 3],
    ...     },
    ...     name="t",
    ... )
    >>> s = pl.Series([9, 5, 4, 1])
    >>> t >> mutate(c=eval_aligned(t.a + s)) >> show()
    Table `t` (backend: polars)
    shape: (4, 3)
    ┌─────┬─────┬─────┐
    │ a   ┆ b   ┆ c   │
    │ --- ┆ --- ┆ --- │
    │ i64 ┆ i64 ┆ i64 │
    ╞═════╪═════╪═════╡
    │ 1   ┆ 2   ┆ 10  │
    │ 2   ┆ 5   ┆ 7   │
    │ 3   ┆ 16  ┆ 7   │
    │ 4   ┆ 3   ┆ 5   │
    └─────┴─────┴─────┘

    Expression containing columns from different tables:

    >>> t1 = pdt.Table({"a": [1, 2, 3, 4]}, name="t1")
    >>> t2 = pdt.Table({"a": [5, 3, 1, 3]}, name="t2")
    >>> t1 >> mutate(c=eval_aligned(t1.a + t2.a, with_=t1)) >> show()
    Table `t1` (backend: polars)
    shape: (4, 2)
    ┌─────┬─────┐
    │ a   ┆ c   │
    │ --- ┆ --- │
    │ i64 ┆ i64 │
    ╞═════╪═════╡
    │ 1   ┆ 6   │
    │ 2   ┆ 5   │
    │ 3   ┆ 4   │
    │ 4   ┆ 7   │
    └─────┴─────┘
    """
    errors.check_arg_type(ColExpr | pl.Series | pd.Series, "eval_aligned", "val", val)
    errors.check_arg_type(Table | Col | None, "eval_aligned", "with_", with_)

    return EvalAligned(val, with_)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import functools
import inspect
import math
import operator
from collections.abc import Iterable, Sequence
from typing import Any
from uuid import UUID

import polars as pl
import sqlalchemy as sqa

from pydiverse.common import (
    Dtype,
    Float,
    Float64,
    Int,
    Int64,
    NullType,
    String,
)
from pydiverse.transform._internal.backend.table_impl import TableImpl
from pydiverse.transform._internal.backend.targets import Polars, SqlAlchemy, Target
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.ops.op import Ftype
from pydiverse.transform._internal.pipe.table import Cache
from pydiverse.transform._internal.tree import types, verbs
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import (
    CaseExpr,
    Cast,
    Col,
    ColExpr,
    ColFn,
    LiteralCol,
    Order,
)
from pydiverse.transform._internal.util.warnings import warn

from . import sqlalchemy_fixes

_ = sqlalchemy_fixes


@dataclasses.dataclass(slots=True)
class Query:
    select: list[UUID]
    partition_by: list[Col] = dataclasses.field(default_factory=list)
    group_by: list[UUID] = dataclasses.field(default_factory=list)
    where: list[ColExpr] = dataclasses.field(default_factory=list)
    having: list[ColExpr] = dataclasses.field(default_factory=list)
    order_by: list[Order] = dataclasses.field(default_factory=list)
    limit: int | None = None
    offset: int | None = None


class SqlImpl(TableImpl):
    def __new__(cls, *args, **kwargs) -> "SqlImpl":
        engine: str | sqa.Engine = inspect.signature(cls.__init__).bind(None, *args, **kwargs).arguments["conf"].engine

        dialect = engine.dialect.name if isinstance(engine, sqa.Engine) else sqa.make_url(engine).get_dialect().name

        # We don't want to import any SQL impls we don't use, so the mapping
        # name -> impl class is defined here.

        if dialect == "sqlite":
            from .sqlite import SqliteImpl

            Impl = SqliteImpl
        elif dialect == "mssql":
            from .mssql import MsSqlImpl

            Impl = MsSqlImpl

        elif dialect == "postgresql":
            from .postgres import PostgresImpl

            Impl = PostgresImpl

        elif dialect == "duckdb":
            from .duckdb import DuckDbImpl

            Impl = DuckDbImpl

        elif dialect == "ibm_db_sa":
            from .ibm_db2 import IbmDb2Impl

            Impl = IbmDb2Impl

        else:
            warn(f"Pydiverse transform is not tested for dialect '{dialect}'. Assuming Postgres compatible SQL.")
            from .postgres import PostgresImpl

            Impl = PostgresImpl

        return super().__new__(Impl)

    def __init__(self, table: str | sqa.Table, conf: SqlAlchemy, name: str | None):
        assert type(self) is not SqlImpl

        self.engine = conf.engine if isinstance(conf.engine, sqa.Engine) else sqa.create_engine(conf.engine)
        if isinstance(table, str):
            self.table = sqa.Table(table, sqa.MetaData(), schema=conf.schema, autoload_with=self.engine)
        else:
            self.table = table

        if name is None:
            name = self.table.name

        super().__init__(
            name,
            {col.name: self.pdt_type(col.type) for col in self.table.columns},
        )

    def _table_def_repr(self) -> str:
        return f"Table('{self.name}', SqlAlchemy('{self.engine.url}'))"

    def col_names(self) -> list[str]:
        return [col.name for col in self.table.columns]

    def schema(self) -> dict[str, Dtype]:
        return {col.name: self.pdt_type(col.type) for col in self.table.columns}

    def _clone(self) -> tuple["SqlImpl", dict[AstNode, AstNode], dict[UUID, UUID]]:
        cloned = self.__class__(self.table, SqlAlchemy(self.engine), self.name)
        return (
            cloned,
            {self: cloned},
            {self.cols[name]._uuid: cloned.cols[name]._uuid for name in self.cols.keys()},
        )

    @classmethod
    def inf(cls):
        return sqa.cast(sqa.literal("inf"), sqa.Double)

    @classmethod
    def nan(cls):
        return sqa.cast(sqa.literal("nan"), sqa.Double)

    @classmethod
    def default_collation(cls) -> str | None:
        return "POSIX"

    @classmethod
    def build_select(cls, nd: AstNode, *, final_select: list[Col] | None = None) -> sqa.Select:
        if final_select is None:
            final_select = Cache.from_ast(nd).selected_cols()
        create_aliases(nd, {})
        nd, query, sqa_expr = cls.compile_ast(nd, {col._uuid: 1 for col in final_select})
        return cls.compile_query(nd, query, sqa_expr)

    @classmethod
    def export(
        cls,
        nd: AstNode,
        target: Target,
        *,
        schema_overrides: dict[UUID, Any],
    ) -> Any:
        final_select = Cache.from_ast(nd).selected_cols()

        sel = cls.build_select(nd, final_select=final_select)
        engine = get_engine(nd)

        if isinstance(target, Polars):
            with engine.connect() as conn:
                df = pl.read_database(
                    sel.compile(engine, compile_kwargs={"literal_binds": True}),
                    connection=conn,
                    schema_overrides={
                        sql_col.name: schema_overrides[col._uuid]
                        for sql_col, col in zip(sel.columns.values(), final_select, strict=True)
                        if col._uuid in schema_overrides
                    }
                    | {
                        sql_col.name: NullType().to_polars()
                        for sql_col, col in zip(sel.columns.values(), final_select, strict=True)
                        if types.without_const(col.dtype()) == NullType()
                    },
                )
                # fix column names in case default case is modified by database driver
                # (for example IBM DB2 returns lowercase columns as uppercase column
                # names)
                df.columns = [c.name for c in sel.columns]
                df.name = nd.name
                return df

        raise NotImplementedError

    @classmethod
    def build_query(cls, nd: AstNode, dialect=None) -> str | None:
        sel = cls.build_select(nd)
        if dialect is None:
            dialect = get_engine(nd).dialect
        return str(sel.compile(dialect=dialect, compile_kwargs={"literal_binds": True}))

    # some backends need to do casting to ensure the correct type
    @classmethod
    def compile_lit(cls, lit: LiteralCol):
        if types.without_const(lit.dtype()).is_float():
            if math.isnan(lit.val):
                return cls.nan()
            elif math.isinf(lit.val):
                return cls.inf() if lit.val > 0 else -cls.inf()

            # Often, floats are incorrectly converted to decimals
            # TODO: ensure in tests that the dtype not only match after export to
            # polars, but also really in the backend
            return sqa.cast(sqa.literal(lit.val, literal_execute=True), cls.sqa_type(lit.dtype()))
        return sqa.literal(lit.val, cls.sqa_type(lit.dtype()), literal_execute=True)

    @classmethod
    def compile_order(cls, order: Order, sqa_expr: dict[str, sqa.Label]) -> sqa.UnaryExpression:
        order_expr = cls.compile_col_expr(order.order_by, sqa_expr)
        if types.without_const(order.order_by.dtype()) == String():
            if cls.default_collation() is not None:
                # SQLAlchemy does not support collations for all backends, so we
                # only apply it if the backend supports it.
                order_expr = order_expr.collate(cls.default_collation())
        order_expr = order_expr.desc() if order.descending else order_expr.asc()
        if order.nulls_last is not None:
            order_expr = order_expr.nulls_last() if order.nulls_last else order_expr.nulls_first()
        return order_expr

    @classmethod
    def compile_cast(cls, cast: Cast, sqa_expr: dict[str, sqa.Label]) -> sqa.Case | sqa.TryCast:
        return cls.cast_compiled(cast, cls.compile_col_expr(cast.val, sqa_expr))

    @classmethod
    def cast_compiled(cls, cast: Cast, compiled_expr: sqa.ColumnElement) -> sqa.Cast | sqa.TryCast:
        if cast.strict:
            return sqa.cast(compiled_expr, cls.sqa_type(cast.target_type))
        else:
            return sqa.try_cast(compiled_expr, cls.sqa_type(cast.target_type))

    @classmethod
    def fix_fn_types(cls, fn: ColFn, val: sqa.ColumnElement, *args: sqa.ColumnElement) -> sqa.ColumnElement:
        return val

    @classmethod
    def compile_ordered_aggregation(
        cls, *args: sqa.ColumnElement, order_by: sqa.UnaryExpression, impl
    ) -> sqa.ColumnElement:
        raise NotImplementedError

    @classmethod
    def compile_col_expr(
        cls, expr: ColExpr, sqa_expr: dict[str, sqa.Label], *, compile_literals=True
    ) -> sqa.ColumnElement:
        if isinstance(expr, Col):
            return sqa_expr[expr._uuid]

        elif isinstance(expr, ColFn):
            args: list[sqa.ColumnElement] = [
                cls.compile_col_expr(arg, sqa_expr, compile_literals=not types.is_const(param))
                for arg, param in zip(
                    expr.args,
                    expr.op.trie.best_match(tuple(arg.dtype() for arg in expr.args))[0],
                    strict=False,
                )
            ]

            assert expr.ftype() is not None

            partition_by = expr.context_kwargs.get("partition_by")
            if partition_by:
                assert expr.ftype() == Ftype.WINDOW
                partition_by = sqa.sql.expression.ClauseList(
                    *(cls.compile_col_expr(col, sqa_expr) for col in partition_by)
                )

            arrange = expr.context_kwargs.get("arrange")
            if arrange:
                order_by = dedup_order_by(cls.compile_order(order, sqa_expr) for order in arrange)
            else:
                order_by = None

            impl = cls.get_impl(expr.op, tuple(arg.dtype() for arg in expr.args))
            impl = functools.partial(impl, _Impl=cls)

            if order_by is not None and expr.ftype() == Ftype.AGGREGATE:
                # some backends need to do preprocessing and some postprocessing here,
                # so we just give them full control by passing the responsibility of
                # calling the `impl`.
                value = cls.compile_ordered_aggregation(*args, order_by=order_by, impl=impl)

            else:
                value: sqa.FunctionElement = impl(*args)

                if expr.op == ops.cum_sum and cls.dialect_order_append_rand():
                    order_by += [cls.get_impl(ops.rand, [])()]

                if partition_by is not None or order_by is not None and expr.ftype() == Ftype.WINDOW:
                    value = sqa.over(
                        value,
                        partition_by=partition_by,
                        order_by=sqa.sql.expression.ClauseList(*order_by) if order_by else None,
                    )

            return cls.fix_fn_types(expr, value, *args)

        elif isinstance(expr, CaseExpr):
            res = sqa.case(
                *(
                    (
                        cls.compile_col_expr(cond, sqa_expr),
                        cls.compile_col_expr(val, sqa_expr),
                    )
                    for cond, val in expr.cases
                ),
                else_=(cls.compile_col_expr(expr.default_val, sqa_expr) if expr.default_val is not None else None),
            )

            if not cls.pdt_type(res.type).is_subtype(expr.dtype()):
                res = res.cast(
                    cls.sqa_type(
                        Int64()
                        if type(expr.dtype()) is Int
                        else Float64()
                        if type(expr.dtype()) is Float
                        else expr.dtype()
                    )
                )

            return res

        elif isinstance(expr, LiteralCol):
            return cls.compile_lit(expr) if compile_literals else expr.val

        elif isinstance(expr, Cast):
            return cls.compile_cast(expr, sqa_expr)

        raise AssertionError

    @classmethod
    def compile_query(cls, table: sqa.Table, query: Query, sqa_expr: dict[UUID, sqa.ColumnElement]) -> sqa.sql.Select:
        sel = table.select().select_from(table)

        if query.where:
            sel = sel.where(*(cls.compile_col_expr(pred, sqa_expr) for pred in query.where))

        if query.group_by:
            sel = sel.group_by(*(sqa_expr[uid] for uid in query.group_by))

        if query.having:
            sel = sel.having(*(cls.compile_col_expr(pred, sqa_expr) for pred in query.having))

        if query.limit is not None:
            sel = sel.limit(query.limit)
            if query.offset:
                sel = sel.offset(query.offset)

        if query.order_by:
            sel = sel.order_by(*dedup_order_by(cls.compile_order(ord, sqa_expr) for ord in query.order_by))

        sel = sel.with_only_columns(*(sqa_expr[uid] for uid in query.select))

        return sel

    @classmethod
    def compile_ast(cls, nd: AstNode, needed_cols: dict[UUID, int]) -> tuple[sqa.Table, Query, dict[UUID, sqa.Label]]:
        if isinstance(nd, verbs.Verb):
            # store a counter in `needed_cols how often each UUID is referenced by
            # ancestors. This allows to only select necessary columns in a subquery.
            for node in nd.iter_col_nodes():
                if isinstance(node, Col):
                    cnt = needed_cols.get(node._uuid)
                    if cnt is None:
                        needed_cols[node._uuid] = 1
                    else:
                        needed_cols[node._uuid] = cnt + 1

            table, query, sqa_expr = cls.compile_ast(nd.child, needed_cols)

        if isinstance(nd, verbs.Mutate | verbs.Summarize):
            query.select = [uid for uid in query.select if sqa_expr[uid].name not in set(nd.names)]

        if isinstance(nd, verbs.SubqueryMarker):
            if needed_cols.keys().isdisjoint(sqa_expr.keys()):
                # We cannot select zero columns from a subquery. This happens when the
                # user only 0-ary functions after the subquery, e.g. `count`.
                needed_cols[query.select[0]] = 1

            # We only want to select those columns that (1) the user uses in some
            # expression later or (2) are present in the final selection.

            original_select = query.select
            query.select = []
            cnt = dict()
            name_in_subquery = dict()

            # resolve potential column name collisions in the subquery
            for uid in needed_cols.keys():
                if uid in sqa_expr:
                    name = sqa_expr[uid].name
                    if c := cnt.get(name):
                        name_in_subquery[uid] = f"{name}_{c}"
                        cnt[name] = c + 1
                    else:
                        name_in_subquery[uid] = name
                        cnt[name] = 1
                    sqa_expr[uid] = sqa.label(name_in_subquery[uid], sqa_expr[uid])
                    query.select.append(uid)

            table = cls.compile_query(table, query, sqa_expr).subquery()
            sqa_expr = {
                uid: sqa.label(name_in_subquery[uid], table.columns.get(name_in_subquery[uid]))
                for uid in needed_cols.keys()
                if uid in sqa_expr
            }

            query = Query(
                [uid for uid in original_select if uid in sqa_expr],
                partition_by=query.partition_by,
            )

        elif isinstance(nd, verbs.Select):
            query.select = [col._uuid for col in nd.select]

        elif isinstance(nd, verbs.Rename):
            sqa_expr = {
                uid: (sqa.label(nd.name_map[lb.name], lb) if lb.name in nd.name_map else lb)
                for uid, lb in sqa_expr.items()
            }

        elif isinstance(nd, verbs.Mutate):
            sqa_expr |= {
                uid: sqa.label(name, cls.compile_col_expr(val, sqa_expr))
                for name, uid, val in zip(nd.names, nd.uuids, nd.values, strict=True)
            }
            query.select += nd.uuids

        elif isinstance(nd, verbs.Filter):
            if query.group_by:
                query.having.extend(nd.predicates)
            else:
                query.where.extend(nd.predicates)

        elif isinstance(nd, verbs.Arrange):
            query.order_by = nd.order_by + query.order_by

        elif isinstance(nd, verbs.Summarize):
            sqa_expr |= {
                uid: sqa.label(name, cls.compile_col_expr(val, sqa_expr))
                for name, uid, val in zip(nd.names, nd.uuids, nd.values, strict=True)
            }
            query.group_by.extend(col._uuid for col in query.partition_by if not types.is_const(col.dtype()))
            query.select = [col._uuid for col in query.partition_by] + nd.uuids
            query.partition_by = []
            query.order_by.clear()

        elif isinstance(nd, verbs.SliceHead):
            if query.limit is None:
                query.limit = nd.n
                query.offset = nd.offset
            else:
                query.limit = min(abs(query.limit - nd.offset), nd.n)
                query.offset += nd.offset

        elif isinstance(nd, verbs.GroupBy):
            if nd.add:
                query.partition_by += nd.group_by
            else:
                query.partition_by = nd.group_by

        elif isinstance(nd, verbs.Ungroup):
            assert not (query.partition_by and query.group_by)
            query.partition_by.clear()

        elif isinstance(nd, verbs.Join):
            right_table, right_query, right_sqa_expr = cls.compile_ast(nd.right, needed_cols)
            sqa_expr.update(right_sqa_expr)

            compiled_on = cls.compile_col_expr(nd.on, sqa_expr)

            if nd.how == "inner":
                query.where.extend(right_query.where)
            elif nd.how == "left":
                compiled_on = functools.reduce(
                    operator.and_,
                    (
                        compiled_on,
                        *(cls.compile_col_expr(pred, right_sqa_expr) for pred in right_query.where),
                    ),
                )
            elif nd.how == "full":
                assert not (query.where or right_query.where)

            table = table.join(
                right_table,
                onclause=compiled_on,
                isouter=nd.how != "inner",
                full=nd.how == "full",
            )

            query.select += right_query.select

            assert not right_query.partition_by
            assert not right_query.group_by

        elif isinstance(nd, verbs.Union):
            # For UNION, both queries must select the same columns in the same order
            # First compile the right AST to get right_query.select (which contains only selected/visible columns)
            right_table, right_query, right_sqa_expr = cls.compile_ast(nd.right, needed_cols)

            # Get column names from both sides
            left_select = query.select
            left_col_names = [sqa_expr[uid].name for uid in left_select]
            right_select = right_query.select
            right_col_names = [right_sqa_expr[uid].name for uid in right_select]

            # If column order doesn't match, wrap right AST with a Select to reorder
            if left_col_names != right_col_names:
                # Get right cache to access Col objects for reordering
                from pydiverse.transform._internal.pipe.cache import Cache

                right_cache = Cache.from_ast(nd.right)

                # Get Col objects from right cache in the order of left columns
                reordered_cols = []
                for name in left_col_names:
                    if name not in right_cache.name_to_uuid:
                        raise ValueError(f"union requires matching column names: '{name}' not found in right table")
                    uid = right_cache.name_to_uuid[name]
                    col = right_cache.cols[uid]
                    reordered_cols.append(col)

                # Wrap right AST with Select to reorder columns and recompile
                right_ast = verbs.Select(nd.right, reordered_cols)
                right_table, right_query, right_sqa_expr = cls.compile_ast(right_ast, needed_cols)

            # Build left and right select statements
            left_sel = cls.compile_query(table, query, sqa_expr)
            right_sel = cls.compile_query(right_table, right_query, right_sqa_expr)

            # If either side is a subquery, get the original CompoundSelect
            # to allow calling sa.union/union_all again
            if isinstance(left_sel, sqa.sql.selectable.Subquery):
                left_sel = left_sel.original
            if isinstance(right_sel, sqa.sql.selectable.Subquery):
                right_sel = right_sel.original

            # Use UNION or UNION ALL
            # distinct=True means UNION (remove duplicates), distinct=False means UNION ALL (keep duplicates)
            if nd.distinct:
                union_query = sqa.union(left_sel, right_sel)
            else:
                union_query = sqa.union_all(left_sel, right_sel)

            # Create a subquery from the union
            table = union_query.subquery()

            # Update sqa_expr to point to the union result columns
            # Use left column names
            sqa_expr = {uid: sqa.label(sqa_expr[uid].name, table.columns[sqa_expr[uid].name]) for uid in left_select}

            # Create a new query with the union result
            # Only keep the select columns, reset all other query state
            query = Query(select=left_select)

            assert not right_query.partition_by
            assert not right_query.group_by

        elif isinstance(nd, TableImpl):
            table = nd.table
            query = Query(select=[col._uuid for col in nd.cols.values()])
            sqa_expr = {
                nd.cols[col.name]._uuid: sqa.label(
                    col.name,
                    sqa.type_coerce(col, cls.sqa_type(nd.cols[col.name].dtype())),
                )
                for col in table.columns
            }

        if isinstance(nd, verbs.Verb):
            # decrease counters (`needed_cols` is not copied)
            for node in nd.iter_col_nodes():
                if isinstance(node, Col):
                    cnt = needed_cols.get(node._uuid)
                    if cnt == 1:
                        del needed_cols[node._uuid]
                    else:
                        needed_cols[node._uuid] = cnt - 1

        return table, query, sqa_expr

    @classmethod
    def sqa_type(cls, pdt_type: Dtype) -> type[sqa.types.TypeEngine]:
        return pdt_type.to_sql()

    @classmethod
    def pdt_type(cls, sqa_type: sqa.types.TypeEngine) -> Dtype:
        return Dtype.from_sql(sqa_type)

    @classmethod
    def dialect_order_append_rand(cls):
        # by default we want to follow polars for cum_sum to ensure equal order values
        # don't get the same result value
        return True


# MSSQL complains about duplicates in ORDER BY.
def dedup_order_by(
    order_by: Iterable[sqa.UnaryExpression],
) -> list[sqa.UnaryExpression]:
    new_order_by: list[sqa.UnaryExpression] = []
    occurred: set[sqa.ColumnElement] = set()

    for ord in order_by:
        peeled = ord
        while isinstance(peeled, sqa.UnaryExpression) and peeled.modifier is not None:
            peeled = peeled.element
        if peeled not in occurred:
            new_order_by.append(ord)
            occurred.add(peeled)

    return new_order_by


# Gives any leaf a unique alias to allow self-joins. We do this here to not force
# the user to come up with dummy names that are not required later anymore. It has
# to be done before a join so that all column references in the join subtrees remain
# valid.
def create_aliases(nd: AstNode, num_occurrences: dict[str, int]) -> dict[str, int]:
    if isinstance(nd, verbs.Verb):
        num_occurrences = create_aliases(nd.child, num_occurrences)

        if isinstance(nd, verbs.Join):
            num_occurrences = create_aliases(nd.right, num_occurrences)

    elif isinstance(nd, TableImpl):
        table_name = nd.table.name
        if cnt := num_occurrences.get(table_name):
            nd.table = nd.table.alias(f"{table_name}_{cnt}")
        else:
            # always set alias to shorten queries with schemas
            nd.table = nd.table.alias(table_name)
            cnt = 0
        num_occurrences[table_name] = cnt + 1

    else:
        raise AssertionError

    return num_occurrences


def get_engine(nd: AstNode) -> sqa.Engine:
    if isinstance(nd, verbs.Verb):
        engine = get_engine(nd.child)

        if isinstance(nd, verbs.Join):
            right_engine = get_engine(nd.right)
            if engine.url != right_engine.url:
                raise NotImplementedError  # TODO: find some good error for this

    else:
        assert isinstance(nd, SqlImpl)
        engine = nd.engine

    return engine


with SqlImpl.impl_store.impl_manager as impl:
    if sqa.__version__ < "2":

        @impl(ops.floordiv)
        def _floordiv(lhs, rhs):
            return sqa.cast(lhs / rhs, sqa.Integer())

    else:

        @impl(ops.floordiv)
        def _floordiv(lhs, rhs):
            return lhs // rhs

    @impl(ops.pow)
    def _pow(lhs, rhs):
        return_type = sqa.Double()
        if isinstance(lhs.type, sqa.Numeric) and isinstance(rhs.type, sqa.Numeric):
            return_type = sqa.Numeric()
        return sqa.func.POW(lhs, rhs, type_=return_type)

    @impl(ops.bool_xor)
    def _xor(lhs, rhs):
        return lhs != rhs

    @impl(ops.pos)
    def _pos(x):
        return x

    @impl(ops.abs)
    def _abs(x):
        return sqa.func.ABS(x, type_=x.type)

    @impl(ops.round)
    def _round(x, decimals=0):
        return sqa.func.ROUND(x, decimals, type_=x.type)

    @impl(ops.is_in)
    def _is_in(x, *values):
        return x.in_(v for v in values)

    @impl(ops.is_null)
    def _is_null(x):
        return x.is_(sqa.null())

    @impl(ops.is_not_null)
    def _is_not_null(x):
        return x.is_not(sqa.null())

    @impl(ops.str_strip)
    def _str_strip(x):
        return sqa.func.TRIM(x, type_=x.type)

    @impl(ops.str_len)
    def _str_length(x):
        return sqa.func.LENGTH(x, type_=sqa.Integer())

    @impl(ops.str_upper)
    def _upper(x):
        return sqa.func.UPPER(x, type_=x.type)

    @impl(ops.str_lower)
    def _lower(x):
        return sqa.func.LOWER(x, type_=x.type)

    @impl(ops.str_replace_all)
    def _str_replace_all(x, y, z):
        return sqa.func.REPLACE(x, y, z, type_=x.type)

    @impl(ops.str_starts_with)
    def _str_starts_with(x, y):
        return x.startswith(y, autoescape=True)

    @impl(ops.str_ends_with)
    def _str_ends_with(x, y):
        return x.endswith(y, autoescape=True)

    @impl(ops.str_slice)
    def _str_slice(x, offset, length):
        # SQL has 1-indexed strings but we do it 0-indexed
        return sqa.func.SUBSTR(x, offset + 1, length)

    @impl(ops.dt_year)
    def _dt_year(x):
        return sqa.extract("year", x)

    @impl(ops.dt_month)
    def _dt_month(x):
        return sqa.extract("month", x)

    @impl(ops.dt_day)
    def _dt_day(x):
        return sqa.extract("day", x)

    @impl(ops.dt_hour)
    def _dt_hour(x):
        return sqa.extract("hour", x)

    @impl(ops.dt_minute)
    def _dt_minute(x):
        return sqa.extract("minute", x)

    @impl(ops.dt_second)
    def _dt_second(x):
        return sqa.extract("second", x)

    @impl(ops.dt_millisecond)
    def _dt_millisecond(x):
        return sqa.extract("millisecond", x) % 1000

    @impl(ops.dt_microsecond)
    def _dt_microsecond(x):
        return sqa.extract("microsecond", x) % 1_000_000

    @impl(ops.dt_day_of_week)
    def _day_of_week(x):
        return sqa.extract("dow", x)

    @impl(ops.dt_day_of_year)
    def _day_of_year(x):
        return sqa.extract("doy", x)

    @impl(ops.horizontal_max)
    def _horizontal_max(*x):
        return sqa.func.GREATEST(*x)

    @impl(ops.horizontal_min)
    def _horizontal_min(*x):
        return sqa.func.LEAST(*x)

    @impl(ops.mean, Int())
    def _mean(x):
        return sqa.func.AVG(x, type_=sqa.Double)

    @impl(ops.mean)
    def _mean(x):
        return sqa.func.AVG(x, type_=x.type)

    @impl(ops.min)
    def _min(x):
        return sqa.func.min(x)

    @impl(ops.max)
    def _max(x):
        return sqa.func.max(x)

    @impl(ops.sum)
    def _sum(x):
        return sqa.func.sum(x)

    @impl(ops.any)
    def _any(x):
        return sqa.func.max(x)

    @impl(ops.all)
    def _all(x):
        return sqa.func.min(x)

    @impl(ops.count)
    def _count(x=None):
        return sqa.func.count(x)

    @impl(ops.count_star)
    def _len():
        return sqa.func.count()

    @impl(ops.shift)
    def _shift(x, by, empty_value=None):
        if by >= 0:
            if empty_value is not None and not isinstance(empty_value.type, sqa.types.NullType):
                return sqa.func.LAG(x, by, empty_value, type_=x.type)
            else:
                return sqa.func.LAG(x, by, type_=x.type)
        if by < 0:
            if empty_value is not None and not isinstance(empty_value.type, sqa.types.NullType):
                return sqa.func.LEAD(x, -by, empty_value, type_=x.type)
            else:
                return sqa.func.LEAD(x, -by, type_=x.type)

    @impl(ops.row_number)
    def _row_number():
        return sqa.func.ROW_NUMBER(type_=sqa.Integer())

    @impl(ops.rank)
    def _rank():
        return sqa.func.rank()

    @impl(ops.dense_rank)
    def _dense_rank():
        return sqa.func.dense_rank()

    @impl(ops.exp)
    def _exp(x):
        return sqa.func.exp(x)

    @impl(ops.log)
    def _log(x):
        return sqa.func.ln(x)

    @impl(ops.floor)
    def _floor(x):
        return sqa.func.floor(x)

    @impl(ops.ceil)
    def _ceil(x):
        return sqa.func.ceil(x)

    @impl(ops.str_to_datetime)
    def _str_to_datetime(x):
        return sqa.cast(x, sqa.DateTime)

    @impl(ops.str_to_date)
    def _str_to_date(x):
        return sqa.cast(x, sqa.Date)

    @impl(ops.is_inf)
    def _is_inf(x, *, _Impl):
        return (x == _Impl.inf()) | (x == -_Impl.inf())

    @impl(ops.is_not_inf)
    def _is_not_inf(x, *, _Impl):
        return (x != _Impl.inf()) & (x != -_Impl.inf())

    @impl(ops.coalesce)
    def _coalesce(*x):
        return sqa.func.coalesce(*x)

    @impl(ops.str_join)
    def _str_join(x, delim):
        return sqa.func.string_agg(x, delim)

    @impl(ops.fill_null)
    def _fill_null(x, y):
        return sqa.func.coalesce(x, y)

    @impl(ops.sin)
    def _sin(x):
        return sqa.func.sin(x)

    @impl(ops.cos)
    def _cos(x):
        return sqa.func.cos(x)

    @impl(ops.tan)
    def _tan(x):
        return sqa.func.tan(x)

    @impl(ops.asin)
    def _asin(x):
        return sqa.func.asin(x)

    @impl(ops.acos)
    def _acos(x):
        return sqa.func.acos(x)

    @impl(ops.atan)
    def _atan(x):
        return sqa.func.atan(x)

    @impl(ops.sqrt)
    def _sqrt(x):
        return sqa.func.sqrt(x)

    @impl(ops.cbrt)
    def _cbrt(x):
        return sqa.func.cbrt(x)

    @impl(ops.log10)
    def _log10(x):
        return sqa.func.log10(x)

    @impl(ops.clip)
    def _clip(x, lower, upper, *, _Impl: SqlImpl, _sig: Sequence[Dtype]):
        return sqa.case(
            (x.is_(sqa.null()), sqa.null()),
            else_=_Impl.get_impl(ops.horizontal_max, _sig)(_Impl.get_impl(ops.horizontal_min, _sig)(x, upper), lower),
        )

    @impl(ops.cum_sum)
    def _cum_sum(x):
        return sqa.func.sum(x)

    @impl(ops.rand)
    def _rand():
        return sqa.func.random()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import inspect
import sys
import warnings as py_warnings

from pydiverse.transform._internal.errors import NonStandardWarning


def warn(
    message: str,
    category: type[Warning] = None,
    stacklevel=1,
):
    py_warnings.warn(message, category, stacklevel=stacklevel + 1)
    return

    stack = inspect.stack(context=0)
    frame = stack[stacklevel]

    # for f in stack[stacklevel:]:
    #     if frame_self := f.frame.f_locals.get("self"):
    #         if isinstance(frame_self, AbstractTableImpl.ExpressionCompiler):
    #             table_impl = frame_self.backend
    #             break

    frame_globals = frame.frame.f_globals
    filename = frame.filename
    lineno = frame.lineno

    del frame  # Prevent reference cycle
    del stack  # Prevent reference cycle

    registry = frame_globals.setdefault("__pydiverse_transform_warnings_registry__", {})
    key = (message, category, lineno)

    if registry.get(key):
        # Not a new warning
        return
    registry[key] = 1

    print(f"{filename}:{lineno}: {category.__name__}: {message}", file=sys.stderr)


def warn_non_standard(
    message: str,
    stacklevel=1,
):
    warn(
        message,
        category=NonStandardWarning,
        stacklevel=stacklevel + 1,
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.transform as pdt
from pydiverse.transform.extended import *
from tests.util import assert_result_equal


def test_ungroup(df3):
    assert_result_equal(df3, lambda t: t >> group_by(t.col1, t.col2) >> ungroup())


def test_select(df3):
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1, t.col2) >> select(t.col1, t.col3),
    )


def test_mutate(df3, df4):
    assert_result_equal(
        df3,
        lambda t: t >> mutate(c1xc2=t.col1 * t.col2) >> group_by(C.c1xc2),
    )

    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1, t.col2) >> mutate(c1xc2=t.col1 * t.col2),
    )

    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> group_by(t.col1, t.col2)
        >> mutate(col1=t.col1 * t.col2)
        >> arrange(t.col3.descending().nulls_last())
        >> ungroup()
        >> left_join(u, t.col2 == u.col2)
        >> mutate(
            x=pdt.row_number(
                arrange=[u.col4.nulls_last(), t.col4.nulls_first()],
                partition_by=[t.col1],
            ),
            p=t.col1 * u.col4,
            y=pdt.rank(arrange=[(t.col1 * u.col4).nulls_last().nulls_first().nulls_last()]),
        ),
    )


def test_grouped_join(df1, df3):
    # Joining a grouped table should always throw an exception
    assert_result_equal(
        (df1, df3),
        lambda t, u: t >> group_by(C.col1) >> join(u, t.col1 == u.col1, how="left"),
        exception=ValueError,
    )

    assert_result_equal(
        (df1, df3),
        lambda t, u: t >> join(u >> group_by(C.col1), t.col1 == u.col1, how="left"),
        exception=ValueError,
    )


@pytest.mark.parametrize("how", ["inner", "left"])
def test_ungrouped_join(df1, df3, how):
    # After ungrouping joining should work again
    assert_result_equal(
        (df1, df3),
        lambda t, u: t >> group_by(t.col1) >> ungroup() >> join(u, t.col1 == u.col1, how=how),
        check_row_order=False,
    )


def test_filter(df3):
    assert_result_equal(df3, lambda t: t >> group_by(t.col1) >> filter(t.col3 >= 2))


def test_arrange(df3):
    assert_result_equal(df3, lambda t: t >> group_by(t.col1) >> arrange(t.col1, -t.col3))

    assert_result_equal(df3, lambda t: t >> group_by(t.col1) >> arrange(-t.col4))


def test_group_by_bool_col(df4):
    assert_result_equal(
        df4,
        lambda t: t >> mutate(x=t.col1 <= t.col2) >> group_by(C.x) >> mutate(y=C.col4.mean()),
    )


def test_group_by_scalar(df3):
    assert_result_equal(df3, lambda t: t >> mutate(x=0) >> group_by(C.x) >> summarize(y=t.col1.sum()))

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(x=0)
        >> mutate(y=C.x.sum(partition_by=t.col2))
        >> group_by(C.y)  # TODO: first alias and then group by should also work
        >> alias()
        >> summarize(z=C.col1.min()),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.transform as pdt
from pydiverse.transform import C
from pydiverse.transform._internal.errors import DataTypeError
from pydiverse.transform._internal.pipe.verbs import mutate
from pydiverse.transform._internal.tree.col_expr import LiteralCol
from tests.fixtures.backend import skip_backends
from tests.util import assert_result_equal


def test_count(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(**{col.name + "_count": col.count() for col in t})
        >> mutate(o=LiteralCol(0).count(filter=t.col3 == 2))
        >> mutate(u=pdt.count(), v=pdt.count(filter=t.col4 > 0)),
    )


def test_row_number(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(row_number=pdt.row_number(arrange=[C.col1.descending().nulls_first(), C.col5.nulls_last()])),
    )


# TODO: add these back in when string comparison is fixed in mssql
@skip_backends("mssql")
def test_min(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            int1=pdt.min(C.col1 + 2, C.col2, 9),
            int2=pdt.min(C.col1 * C.col2, 0),
            int3=pdt.min(C.col1 * C.col2, C.col2 * C.col3, 2 - C.col3),
            int4=pdt.min(C.col1),
            float1=pdt.min(C.col1, 1.5),
            float2=pdt.min(1, C.col1 + 1.5, C.col2, 2.2),
            str1=pdt.min(C.col5, "c"),
            str2=pdt.min(C.col5, "C"),
        ),
    )


@skip_backends("mssql")
def test_max(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            int1=pdt.max(C.col1 + 2, C.col2, 9),
            int2=pdt.max(C.col1 * C.col2, 0),
            int3=pdt.max(C.col1 * C.col2, C.col2 * C.col3, 2 - C.col3),
            int4=pdt.max(C.col1),
            float1=pdt.max(C.col1, 1.5),
            float2=pdt.max(1, C.col1 + 1.5, C.col2, 2.2),
            str1=pdt.max(C.col5, "c"),
            str2=pdt.max(C.col5, "C"),
        ),
    )


def test_any_all(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            g=pdt.any(t.col1 == 0, t.col2 == 0),
            h=pdt.all(t.col1 < t.col4, t.col4 > 0, t.col5.str.len() <= 1),
        ),
    )


def test_sum(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            w=pdt.min(*(c for c in t if c.dtype().is_int())),
            u=pdt.sum(*(c for c in t if c.dtype().is_int())),
            h=pdt.sum(t.col1 * t.col4, t.col5.str.len() * 2, t.col2 - t.col1),
        ),
    )


def test_coalesce(df4):
    assert_result_equal(
        df4,
        lambda t: t >> mutate(w=pdt.coalesce(*(c for c in t))),
        exception=DataTypeError,
    )

    assert_result_equal(
        df4,
        lambda t: t >> mutate(w=pdt.coalesce(*(c for c in t if c.dtype().is_int()))),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any
from uuid import UUID

import duckdb
import duckdb_engine
import polars as pl
import sqlalchemy as sqa

from pydiverse.common import Dtype
from pydiverse.transform._internal.backend.duckdb import DuckDbImpl
from pydiverse.transform._internal.backend.table_impl import TableImpl
from pydiverse.transform._internal.backend.targets import Polars, Target
from pydiverse.transform._internal.tree.ast import AstNode


# TODO: we should move the engine of SqlImpl in the subclasses and let this thing
# inherit from SqlImpl in order to make the usage of SqlImpl.compile_ast more clean.
# Currently it works only since this class also has a table object, but it should be
# enforced by inheritance.
class DuckDbPolarsImpl(TableImpl):
    backend_name = "duckdb_polars"

    def __init__(self, name: str | None, df: pl.DataFrame | pl.LazyFrame):
        self.df = df if isinstance(df, pl.LazyFrame) else df.lazy()

        super().__init__(
            name,
            {name: Dtype.from_polars(dtype) for name, dtype in df.collect_schema().items()},
        )

        self.table = sqa.Table(
            name or "<polars dataframe>",
            sqa.MetaData(),
            *(sqa.Column(col.name, DuckDbImpl.sqa_type(col.dtype())) for col in self.cols.values()),
        )

    def _table_def_repr(self) -> str:
        return f"Table(df, DuckDb(), name='{self.name}')"

    @staticmethod
    def build_query(nd: AstNode) -> str | None:
        return DuckDbImpl.build_query(nd, dialect=duckdb_engine.Dialect())

    @staticmethod
    def export(
        nd: AstNode,
        target: Target,
        *,
        schema_overrides: dict[UUID, Any],  # TODO: use this
    ) -> pl.DataFrame:
        if isinstance(target, Polars):
            sel = DuckDbImpl.build_select(nd)
            query_str = str(
                sel.compile(
                    dialect=duckdb_engine.Dialect(),
                    compile_kwargs={"literal_binds": True},
                )
            )

            # tell duckdb which table names in the SQL query correspond to which
            # data frames
            for desc in nd.iter_subtree_postorder():
                if isinstance(desc, DuckDbPolarsImpl):
                    duckdb.register(desc.table.name, desc.df)

            return duckdb.sql(query_str).pl()

        raise AssertionError

    def _clone(self) -> tuple[AstNode, dict[AstNode, AstNode], dict[UUID, UUID]]:
        cloned = DuckDbPolarsImpl(self.name, self.df)
        return (
            cloned,
            {self: cloned},
            {self.cols[name]._uuid: cloned.cols[name]._uuid for name in self.cols.keys()},
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import polars as pl
import pytest
import sqlalchemy as sqa

import pydiverse.transform as pdt
from pydiverse.transform.extended import *
from tests.util import assert_equal

df1 = pl.DataFrame(
    {
        "col1": [1, 2, 3, 4],
        "col2": ["a", "b", "c", "d"],
    }
)

df2 = pl.DataFrame(
    {
        "col1": [1, 2, 2, 4, 5, 6],
        "col2": [2, 2, 0, 0, 2, None],
        "col3": [0.0, 0.1, 0.2, 0.3, 0.01, 0.02],
    }
)

df3 = pl.DataFrame(
    {
        "col1": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2],
        "col2": [0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
        "col3": [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        "col4": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        "col5": list("abcdefghijkl"),
    }
)

df4 = pl.DataFrame(
    {
        "col1": [None, 0, 0, 0, 0, None, 1, 1, 1, 2, 2, 2, 2],
        "col2": [0, 0, 1, 1, 0, 0, 1, None, 1, 0, 0, 1, 1],
        "col3": [None, None, None, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        "col4": [None, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        "col5": list("abcdefghijkl") + [None],
    }
)

df_left = pl.DataFrame(
    {
        "a": [1, 2, 3, 4],
    }
)

df_right = pl.DataFrame(
    {
        "b": [0, 1, 2, 2],
        "c": [5, 6, 7, 8],
    }
)


@pytest.fixture
def engine():
    engine = sqa.create_engine("sqlite:///:memory:")
    # engine = sqa.create_engine("postgresql://sqa:Pydiverse23@127.0.0.1:6543")
    # engine = sqa.create_engine(
    #     "mssql+pyodbc://sqa:PydiQuant27@127.0.0.1:1433"
    #     "/master?driver=ODBC+Driver+18+for+SQL+Server&encrypt=no"
    # )
    # engine = sqa.create_engine("db2+ibm_db://db2inst1:password@localhost:50000/testdb")

    df1.write_database("df1", engine, if_table_exists="replace")
    df2.write_database("df2", engine, if_table_exists="replace")
    df3.write_database("df3", engine, if_table_exists="replace")
    df4.write_database("df4", engine, if_table_exists="replace")
    df_left.write_database("df_left", engine, if_table_exists="replace")
    df_right.write_database("df_right", engine, if_table_exists="replace")
    return engine


@pytest.fixture
def tbl1(engine):
    return pdt.Table("df1", SqlAlchemy(engine))


@pytest.fixture
def tbl2(engine):
    return pdt.Table("df2", SqlAlchemy(engine))


@pytest.fixture
def tbl3(engine):
    return pdt.Table("df3", SqlAlchemy(engine))


@pytest.fixture
def tbl4(engine):
    return pdt.Table("df4", SqlAlchemy(engine))


@pytest.fixture
def tbl_left(engine):
    return pdt.Table("df_left", SqlAlchemy(engine))


@pytest.fixture
def tbl_right(engine):
    return pdt.Table("df_right", SqlAlchemy(engine))


class TestSqlTable:
    def test_build_query(self, tbl1):
        query_str = tbl1 >> build_query()
        expected_out = "SELECT df1.col1 AS col1, df1.col2 AS col2 FROM df1 as df1"
        assert query_str.lower().split() == expected_out.lower().split()

    def test_show_query(self, tbl1, capfd):
        tbl1 >> show_query()
        out = capfd.readouterr().out
        expected_out = "SELECT df1.col1 AS col1, df1.col2 AS col2 FROM df1 as df1"

        assert out.lower().split() == expected_out.lower().split()

        # Verify that it is chainable
        tbl1 >> show_query(pipe=True) >> collect()

    def test_export(self, tbl1):
        assert_equal(tbl1 >> export(Polars()), df1)

    def test_select(self, tbl1):
        assert_equal(tbl1 >> select(tbl1.col1), df1.select("col1"))
        assert_equal(tbl1 >> select(tbl1.col2), df1.select("col2"))

    def test_mutate(self, tbl1):
        assert_equal(
            tbl1 >> mutate(col1times2=tbl1.col1 * 2),
            pl.DataFrame(
                {
                    "col1": [1, 2, 3, 4],
                    "col2": ["a", "b", "c", "d"],
                    "col1times2": [2, 4, 6, 8],
                }
            ),
        )

        assert_equal(
            tbl1 >> mutate(col1times2=tbl1.col1 * 2) >> select(C.col1times2),
            pl.DataFrame(
                {
                    "col1times2": [2, 4, 6, 8],
                }
            ),
        )

        # Check proper column referencing
        t = tbl1 >> mutate(col2=tbl1.col1, col1=tbl1.col2)
        assert_equal(
            t >> select() >> mutate(x=t.col1, y=t.col2),
            tbl1 >> select() >> mutate(x=tbl1.col2, y=tbl1.col1),
        )
        assert_equal(
            t >> select() >> mutate(x=tbl1.col1, y=tbl1.col2),
            tbl1 >> select() >> mutate(x=tbl1.col1, y=tbl1.col2),
        )

    def test_join(self, tbl_left, tbl_right):
        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "left", suffix="")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame({"a": [1, 2, 2, 3, 4], "b": [1, 2, 2, None, None]}),
            check_row_order=False,
        )

        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "inner", suffix="")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame({"a": [1, 2, 2], "b": [1, 2, 2]}),
        )

        assert_equal(
            (
                tbl_left
                >> join(tbl_right, tbl_left.a == tbl_right.b, "full", suffix="_1729")
                >> select(tbl_left.a, tbl_right.b)
            ),
            pl.DataFrame(
                {
                    "a": [1, 2, 2, 3, 4, None],
                    "b_1729": [1, 2, 2, None, None, 0],
                }
            ),
            check_row_order=False,
        )

    def test_filter(self, tbl1):
        # Simple filter expressions
        assert_equal(tbl1 >> filter(tbl1.col1 == tbl1.col1), df1)
        assert_equal(tbl1 >> filter(tbl1.col1 == 3), df1.filter(pl.col("col1") == 3))

        # More complex expressions
        assert_equal(
            tbl1 >> filter(tbl1.col1 // 2 == 1),
            pl.DataFrame({"col1": [2, 3], "col2": ["b", "c"]}),
        )

        assert_equal(
            tbl1 >> filter(1 < tbl1.col1) >> filter(tbl1.col1 < 4),
            df1.filter((1 < pl.col("col1")) & (pl.col("col1") < 4)),
        )

    def test_arrange(self, tbl2):
        assert_equal(
            tbl2 >> arrange(tbl2.col3) >> select(tbl2.col3),
            df2.select("col3").sort("col3", descending=False),
        )

        assert_equal(
            tbl2 >> arrange(-tbl2.col3) >> select(tbl2.col3),
            df2.select("col3").sort("col3", descending=True),
        )

        assert_equal(
            tbl2 >> arrange(tbl2.col1, tbl2.col2),
            df2.sort(["col1", "col2"], descending=[False, False]),
        )

        assert_equal(
            tbl2 >> arrange(tbl2.col1, tbl2.col2),
            tbl2 >> arrange(tbl2.col2) >> arrange(tbl2.col1),
        )

        assert_equal(tbl2 >> arrange(--tbl2.col3), tbl2 >> arrange(tbl2.col3))  # noqa: B002

    def test_summarize(self, tbl3):
        assert_equal(
            tbl3 >> summarize(mean=tbl3.col1.mean(), max=tbl3.col4.max()),
            pl.DataFrame({"mean": [1], "max": [11]}),
        )

        assert_equal(
            tbl3 >> group_by(tbl3.col1) >> summarize(mean=tbl3.col4.mean()),
            pl.DataFrame({"col1": [0, 1, 2], "mean": [1.5, 5.5, 9.5]}),
            check_row_order=False,
        )

        assert_equal(
            tbl3 >> summarize(mean=tbl3.col4.mean()) >> mutate(mean_2x=C.mean * 2),
            pl.DataFrame({"mean": [5.5], "mean_2x": [11]}),
        )

    def test_group_by(self, tbl3):
        # Grouping doesn't change the result
        assert_equal(tbl3 >> group_by(tbl3.col1), tbl3)
        assert_equal(
            tbl3 >> summarize(mean4=tbl3.col4.mean()) >> group_by(C.mean4),
            tbl3 >> summarize(mean4=tbl3.col4.mean()),
        )

        # Groupings can be added
        assert_equal(
            tbl3
            >> group_by(tbl3.col1)
            >> group_by(tbl3.col2, add=True)
            >> summarize(mean3=tbl3.col3.mean(), mean4=tbl3.col4.mean()),
            tbl3 >> group_by(tbl3.col1, tbl3.col2) >> summarize(mean3=tbl3.col3.mean(), mean4=tbl3.col4.mean()),
        )

        # Ungroup doesn't change the result
        assert_equal(
            tbl3 >> group_by(tbl3.col1) >> summarize(mean4=tbl3.col4.mean()) >> ungroup(),
            tbl3 >> group_by(tbl3.col1) >> summarize(mean4=tbl3.col4.mean()),
        )

    def test_alias(self, tbl1, tbl2):
        x = tbl2 >> alias("x")
        assert x._ast.name == "x"

        # Check that applying alias doesn't change the output
        a = (
            tbl1
            >> mutate(xyz=(tbl1.col1 * tbl1.col1) // 2)
            >> join(tbl2, tbl1.col1 == tbl2.col1, "left")
            >> mutate(col1=tbl1.col1 - C.xyz)
        )
        b = a >> alias("b")

        assert_equal(a, b)

        # Self Join
        self_join = tbl2 >> join(x, tbl2.col1 == x.col1, "left", suffix="42") >> alias("self_join")

        self_join_expected = df2.join(
            df2,
            how="left",
            left_on="col1",
            right_on="col1",
            coalesce=False,
            suffix="42",
        )

        assert_equal(self_join, self_join_expected, check_row_order=False)

    def test_lambda_column(self, tbl1, tbl2):
        # Select
        assert_equal(tbl1 >> select(C.col1), tbl1 >> select(tbl1.col1))

        # Mutate
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> mutate(b=C.a * 2) >> select(C.b),
            tbl1 >> select() >> mutate(b=tbl1.col1 * 4),
        )

        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> mutate(b=C.a * 2, a=tbl1.col1) >> select(C.b),
            tbl1 >> select() >> mutate(b=tbl1.col1 * 4),
        )

        # Join
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1) >> join(tbl2, C.a == tbl2.col1, "left") >> select(C.a, *tbl2),
            tbl1 >> select() >> mutate(a=tbl1.col1) >> join(tbl2, tbl1.col1 == tbl2.col1, "left", suffix="_df2"),
        )

        # Filter
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> filter(C.a % 2 == 0),
            tbl1 >> mutate(a=tbl1.col1 * 2) >> filter((tbl1.col1 * 2) % 2 == 0),
        )

        # Arrange
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> arrange(C.a),
            tbl1 >> arrange(tbl1.col1) >> mutate(a=tbl1.col1 * 2),
        )

    def test_select_without_tbl_ref(self, tbl2):
        assert_equal(
            tbl2 >> summarize(count=pdt.count()),
            tbl2 >> summarize(count=tbl2.col1.count()),
        )

        assert_equal(tbl2 >> summarize(count=count()), pl.DataFrame({"count": [len(df2)]}))

    def test_null_comparison(self, tbl4):
        assert_equal(
            tbl4 >> mutate(u=tbl4.col1 == tbl4.col3),
            df4.with_columns((pl.col("col1") == pl.col("col3")).alias("u")),
        )

        assert_equal(
            tbl4 >> mutate(u=tbl4.col3.is_null()),
            df4.with_columns(pl.col("col3").is_null().alias("u")),
        )

    def test_case_expression(self, tbl3):
        assert_equal(
            (
                tbl3
                >> mutate(
                    col1=when(C.col1 == 0).then(1).when(C.col1 == 1).then(2).when(C.col1 == 2).then(3).otherwise(-1)
                )
                >> select(C.col1)
            ),
            (df3.select("col1") + 1),
        )

        assert_equal(
            (
                tbl3
                >> mutate(x=when(C.col1 == C.col2).then(1).when(C.col1 == C.col3).then(2).otherwise(C.col4))
                >> select(C.x)
            ),
            pl.DataFrame({"x": [1, 1, 2, 3, 4, 2, 1, 1, 8, 9, 2, 11]}),
        )

    def test_ast_repr(self, tbl3):
        tbl3 >> ast_repr()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform.extended import *
from tests.util.assertion import assert_result_equal


def test_dtypes(df1):
    assert_result_equal(
        df1,
        lambda t: t >> filter(t.col1 % 2 == 1) >> inner_join(s := t >> mutate(u=t.col1 % 2) >> alias(), t.col1 == s.u),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from pydiverse.transform._internal.errors import FunctionTypeError
from pydiverse.transform.extended import *
from tests.util import assert_result_equal


def test_alias(df3):
    assert_result_equal(df3, lambda t: t >> alias())


def test_alias_named(df3):
    def run(post_op):
        assert_result_equal(df3, lambda t: t >> alias("a") >> post_op)

    for post_op in [
        mutate(),
        mutate(y=C.col1),
        mutate(y=C.col1.count()),
        mutate(y=C.col1.count(partition_by=C.col2)),
        summarize(y=C.col1.count()),
    ]:
        run(post_op)


def test_alias_mutate(df3):
    def run(post_op):
        assert_result_equal(df3, lambda t: t >> mutate(x=t.col1) >> alias("a") >> post_op)

    for post_op in [
        mutate(),
        mutate(y=C.col1),
        mutate(y=C.col1.count()),
        mutate(y=C.col1.count(partition_by=C.col2)),
        summarize(y=C.col1.count()),
    ]:
        run(post_op)


def test_alias_window(df3):
    def run(post_op):
        assert_result_equal(
            df3,
            lambda t: t >> mutate(x=t.col1.count(partition_by=t.col2)) >> alias("a") >> post_op,
        )

    for post_op in [
        mutate(),
        mutate(y=C.col1),
        mutate(y=C.col1.count()),
        mutate(y=C.col1.count(partition_by=C.col2)),
        summarize(y=C.col1.count() + C.x.sum()),
    ]:
        run(post_op)

    with pytest.raises(FunctionTypeError):
        run(summarize(y=C.col1.count() + C.x))


@pytest.mark.xfail  # TODO: fix problem with errors
def test_alias_summarize(df3):
    def run(post_op):
        assert_result_equal(
            df3,
            lambda t: t >> group_by(t.col1, t.col2) >> summarize(x=t.col1.count()) >> alias("a") >> post_op,
        )

    for post_op in [
        mutate(),
        mutate(y=C.col1),
        mutate(y=C.col1.count()),
        mutate(y=C.col1.count(partition_by=C.col2)),
        summarize(y=C.col1.count()),
        mutate(y=C.x < row_number()),
    ]:
        run(post_op)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.transform as pdt
from pydiverse.common import Float32, Float64, Int32, Int64
from pydiverse.transform._internal.pipe.c import C
from pydiverse.transform._internal.pipe.verbs import mutate
from tests.util.assertion import assert_result_equal


def test_string_to_float(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t >> mutate(u=t.c.cast(pdt.Float64())),
    )


def test_string_to_int(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t >> mutate(u=t.d.cast(pdt.Int64())),
    )


def test_float_to_int(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(
            **{
                col.name: pdt.when((col <= (1 << 63) - 1) & (col >= -(1 << 63)))
                .then(col)
                .otherwise(0)
                .cast(pdt.Float64())
                for col in t
            }
        ),
    )

    # all backends throw on out of range values
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.cast(pdt.Int64()) for c in t}),
        exception=Exception,
    )


def test_datetime_to_date(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t >> mutate(u=t.col1.cast(pdt.Date()), v=t.col2.cast(pdt.Date())),
    )


def test_int_to_string(df_int):
    assert_result_equal(df_int, lambda t: t >> mutate(**{c.name: c.cast(pdt.String()) for c in t}))


def test_float_to_string(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(**{c.name: c.cast(pdt.String()) for c in t})
        >> (lambda s: mutate(**{c.name: c.cast(pdt.Float64()) for c in s})),
    )


def test_datetime_to_string(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=t.col1.cast(pdt.String()),
            y=t.col2.cast(pdt.String()),
        )
        >> mutate(
            x=C.x.str.to_datetime(),
            y=C.y.str.to_datetime(),
        ),
    )


def test_date_to_string(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=t.col1.cast(pdt.Date()).cast(pdt.String()),
            y=t.col2.cast(pdt.Date()).cast(pdt.String()),
            z=t.cdate.cast(pdt.String()),
        ),
    )


@pytest.mark.xfail
def test_non_strict_cast(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            **{
                f"{col.name}_{dtype}": col.cast(dtype, strict=False)
                for col in t
                for dtype in (Int64(), Int32(), Float64(), Float32())
            }
        ),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import inspect
from collections.abc import Callable, Iterable
from html import escape
from typing import Literal

import pandas as pd
import polars as pl
import sqlalchemy as sqa

from pydiverse.transform._internal import errors
from pydiverse.transform._internal.backend.table_impl import TableImpl
from pydiverse.transform._internal.backend.targets import Target
from pydiverse.transform._internal.errors import ColumnNotFoundError
from pydiverse.transform._internal.pipe.cache import Cache
from pydiverse.transform._internal.pipe.pipeable import Pipeable
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import Col, ColName


class Table:
    __slots__ = ["_ast", "_cache"]

    """
    All attributes of a table are columns except for the `_ast` attribute
    which is a reference to the underlying abstract syntax tree.
    """

    def __init__(
        self,
        resource: pl.DataFrame | pl.LazyFrame | pd.DataFrame | sqa.Table | str | dict,
        backend: Target | None = None,
        *,
        name: str | None = None,
    ):
        """
        Creates a new table.

        :param resource:
            The data source to construct the table from. This can be a polars or pandas
            data frame, a python dictionary, a SQLAlchemy table or the name of a table
            in a SQL database.

        :param backend:
            The execution backend. This must be one of the pydiverse.transform backend
            objects, see :doc:`targets`. It may carry additional information how to
            interpret the *resource* argument, such as a SQLAlchemy engine.

        :param name:
            The name of the table. It is not required to give the table a name, but may
            make print output more readable.

        Examples
        --------
        **Python dictionary**.

        >>> t = pdt.Table(
        ...     {
        ...         "a": [4, 3, -35, 24, 105],
        ...         "b": [4, 4, 0, -23, 42],
        ...     },
        ...     name="T",
        ... )
        >>> t >> show()
        Table T, backend: PolarsImpl
        shape: (5, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 4   ┆ 4   │
        │ 3   ┆ 4   │
        │ -35 ┆ 0   │
        │ 24  ┆ -23 │
        │ 105 ┆ 42  │
        └─────┴─────┘

        **Polars data frame.**

        >>> df = pl.DataFrame(
        ...     {
        ...         "a": [4, 3, -35, 24, 105],
        ...         "b": ["a", "o", "---", "i23", "  "],
        ...     },
        ... )
        >>> t = pdt.Table(df, name="T")
        >>> t >> show()
        Table T, backend: PolarsImpl
        shape: (5, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 4   ┆ a   │
        │ 3   ┆ o   │
        │ -35 ┆ --- │
        │ 24  ┆ i23 │
        │ 105 ┆     │
        └─────┴─────┘

        **Pandas data frame.** Note that the data frame is converted to a polars data
        frame and the backend is polars.

        >>> import pandas as pd
        >>> df = pd.DataFrame(
        ...     {
        ...         "a": [4, 3, -35, 24, 105],
        ...         "b": ["a", "o", "---", "i23", "  "],
        ...     },
        ... )
        >>> t = pdt.Table(df, name="T")
        >>> t >> show()
        Table T, backend: PolarsImpl
        shape: (5, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 4   ┆ a   │
        │ 3   ┆ o   │
        │ -35 ┆ --- │
        │ 24  ┆ i23 │
        │ 105 ┆     │
        └─────┴─────┘

        **SQL.** Assuming you have a SQLAlchemy engine ``engine``, which is has a
        connection to a database containing a table ``t1`` in a schema ``s1``, you can
        create a pydiverse.transform Table from it as follows.

        >>> t = pdt.Table("t1", SqlAlchemy(engine, schema="s1"))
        >>> t >> show()
        Table t1, backend: PostgresImpl
        shape: (5, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 4   ┆ a   │
        │ 3   ┆ o   │
        │ -35 ┆ --- │
        │ 24  ┆ i23 │
        │ 105 ┆     │
        └─────┴─────┘

        Note that the name argument to the ``pdt.Table`` constructor was not specified,
        so transform used the name of the SQL table. This example of course assumes that
        a database connection is set up and the above table is already present in the
        database. For more information on how to set up a connection, see
        :doc:`/database_testing`.
        """

        # AstNodes are also allowed for `resource`, but we do not expose this to the
        # user.
        if not isinstance(resource, AstNode):
            errors.check_arg_type(
                pl.DataFrame | pl.LazyFrame | pd.DataFrame | sqa.Table | str | dict,
                "Table.__init__",
                "resource",
                resource,
            )
        errors.check_arg_type(Target | None, "Table.__init__", "backend", backend)
        errors.check_arg_type(str | None, "Table.__init__", "name", name)

        self._ast: AstNode = TableImpl.from_resource(resource, backend, name=name)
        self._cache = Cache.from_ast(self._ast)

    def __getitem__(self, key: str | Col | ColName) -> Col:
        errors.check_arg_type(str | Col | ColName, "Table.__getitem__", "key", key)
        if isinstance(key, ColName):
            key = key.name
        elif isinstance(key, Col):
            # This functionality is useful to get the name of a column in a table via a
            # reference of a previous table.
            if key._uuid not in self._cache.uuid_to_name:
                raise ColumnNotFoundError(
                    f"column `{key.ast_repr()}` does not exist in table `{self._ast.short_name()}`"
                )
            return Col(
                self._cache.uuid_to_name[key._uuid],
                self._ast,
                key._uuid,
                key._dtype,
                key._ftype,
            )
        return self.__getattr__(key)

    def __getattr__(self, name: str) -> Col:
        if name in ("__copy__", "__deepcopy__", "__setstate__", "__getstate__"):
            # for hasattr to work correctly on dunder methods
            raise AttributeError
        if name not in self._cache.name_to_uuid:
            raise ColumnNotFoundError(f"column `{name}` does not exist in table `{self._ast.short_name()}`")
        col = self._cache.cols[self._cache.name_to_uuid[name]]
        return Col(name, self._ast, col._uuid, col._dtype, col._ftype)

    def __setstate__(self, d):  # to avoid very annoying AttributeErrors
        for slot, val in d[1].items():
            setattr(self, slot, val)

    def __iter__(self) -> Iterable[Col]:
        cols = [self.__getattr__(name) for name in self._cache.name_to_uuid.keys()]
        yield from cols

    def __contains__(self, col: str | Col | ColName) -> bool:
        if isinstance(col, Col):
            return col._uuid in self._cache.uuid_to_name
        if isinstance(col, ColName):
            col = col.name
        return col in self._cache.name_to_uuid

    def __len__(self) -> int:
        return len(self._cache.name_to_uuid)

    def __rshift__(self, rhs) -> "Table":
        """
        The pipe operator for chaining verbs.
        """

        if isinstance(rhs, Pipeable):
            return rhs(self)
        if isinstance(rhs, Callable):
            num_params = len(inspect.signature(rhs).parameters)
            if num_params != 1:
                raise TypeError(
                    "only functions with one parameter can be used in a pipe, got "
                    f"function with {num_params} parameters."
                )
            return self >> rhs(self)

        raise TypeError(
            f"found instance of invalid type `{type(rhs)}` in the pipe. \n"
            "hint: You can use a `Table` or a Callable taking a single argument in a "
            "pipe. If you use a Callable, it will receive the current table as an "
            "and must return a `Table`."
        )

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        from pydiverse.transform._internal.pipe.verbs import (
            build_query,
            name,
        )

        if self >> name() is None:
            tbl_name = "Table without name "
        else:
            tbl_name = f"Table `{self >> name()}` "

        res = tbl_name + f"(backend: {self._cache.backend.backend_name})\n"

        try:
            df, height = get_head_tail(self)
            res += f"shape: ({height}, {len(self)})\n"

        except Exception as e:
            try:
                ast = repr(self._ast)
            except Exception:
                ast = ""
            return res + f"export failed\n{type(e).__name__}: {str(e)}\n{ast}"

        try:
            query = self >> build_query()
            if query is None:
                query = ""
            else:
                query = "\n\nQuery:\n" + query
        except Exception:
            query = ""
        return res + str(df).split("\n", 1)[1]

    def __dir__(self) -> list[str]:
        return [name for name in self._cache.name_to_uuid.keys()]

    def _repr_html_(self) -> str:
        from pydiverse.transform._internal.pipe.verbs import name

        if self >> name() is None:
            tbl_name = "Table without name "
        else:
            tbl_name = f"Table <code>{self >> name()}</code> "

        html = tbl_name + f"(backend: <code>{self._cache.backend.backend_name}</code>)</br>"
        try:
            # We use polars' _repr_html_ on the first and last few rows of the table and
            # fix the `shape` afterwards.
            df, height = get_head_tail(self)
            html += f"<small>shape: ({height}, {len(self)})</small>"

            df_html = df._repr_html_()
            num_rows_begin = df_html.find("shape: (")
            num_rows_end = df_html.find(",", num_rows_begin)

        except Exception as e:
            return html + (f"</br><pre>export failed\n{escape(e.__class__.__name__)}: {escape(str(e))}</pre>")

        return f"{df_html[: num_rows_begin + 8]}{height}{df_html[num_rows_end:]}"

    def _repr_pretty_(self, p, cycle):
        p.text(str(self) if not cycle else "...")


def get_head_tail(tbl: Table) -> tuple[pl.DataFrame, int]:
    import pydiverse.transform as pdt
    from pydiverse.transform._internal.pipe.verbs import export, slice_head, summarize

    height = tbl >> summarize(num_rows=pdt.count()) >> export(pdt.Scalar)
    tbl_rows = int(pl.Config.state().get("POLARS_FMT_MAX_ROWS") or 10)
    head_tail_len = tbl_rows // 2 + 1

    # Only export the first and last few rows.
    head: pl.DataFrame = tbl >> slice_head(head_tail_len) >> export(pdt.Polars)
    tail: pl.DataFrame = (
        tbl >> slice_head(head_tail_len, offset=max(head_tail_len, height - head_tail_len)) >> export(pdt.Polars)
    )
    return pl.concat([head, tail]), height


def backend(
    table: Table,
) -> Literal["polars", "duckdb_polars", "sqlite", "postgres", "duckdb", "mssql", "ibm_db2"]:
    """
    Returns the backend of the table as a string.
    """
    return table._cache.backend.backend_name


def is_sql_backed(table: Table) -> bool:
    """
    Whether the table has a SQL backend.
    """
    return not backend(table).endswith("polars")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import sqlalchemy as sqa

from pydiverse.common import (
    Date,
    Datetime,
    Float,
    String,
)
from pydiverse.transform._internal.backend.sql import SqlImpl
from pydiverse.transform._internal.errors import NotSupportedError
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.tree import types
from pydiverse.transform._internal.tree.col_expr import Cast, ColFn
from pydiverse.transform._internal.util.warnings import warn_non_standard


class SqliteImpl(SqlImpl):
    backend_name = "sqlite"

    @classmethod
    def inf(cls):
        return sqa.type_coerce(sqa.literal("1e314"), sqa.Double)

    @classmethod
    def nan(cls):
        raise NotSupportedError("SQLite does not have `nan`, use `null` instead.")

    @classmethod
    def default_collation(cls):
        return "BINARY"

    @classmethod
    def compile_cast(cls, cast: Cast, sqa_col: dict[str, sqa.Label]) -> sqa.Cast:
        compiled_val = cls.compile_col_expr(cast.val, sqa_col)
        val_type = types.without_const(cast.val.dtype())

        if val_type == String() and cast.target_type.is_float():
            return sqa.case(
                (compiled_val == "inf", cls.inf()),
                (compiled_val == "-inf", -cls.inf()),
                else_=cls.cast_compiled(cast, compiled_val),
            )

        elif val_type == Datetime() and cast.target_type == Date():
            return sqa.type_coerce(sqa.func.date(compiled_val), sqa.Date())
        elif val_type == Date() and cast.target_type == Datetime():
            return sqa.type_coerce(sqa.func.datetime(compiled_val), sqa.DateTime())

        elif val_type.is_float() and cast.target_type == String():
            return sqa.case(
                (compiled_val == cls.inf(), "inf"),
                (compiled_val == -cls.inf(), "-inf"),
                else_=cls.cast_compiled(cast, compiled_val),
            )

        return cls.cast_compiled(cast, compiled_val)

    @classmethod
    def cast_compiled(cls, cast: Cast, compiled_expr: sqa.ColumnElement):
        # For SQLite, we ignore the `strict` parameter to `cast`.
        return sqa.cast(compiled_expr, cls.sqa_type(cast.target_type))

    @classmethod
    def fix_fn_types(cls, fn: ColFn, val: sqa.ColumnElement, *args: sqa.ColumnElement) -> sqa.ColumnElement:
        if fn.op in (ops.horizontal_min, ops.horizontal_max, ops.mean, ops.min, ops.max) and fn.dtype().is_float():
            return sqa.cast(val, sqa.Double)
        return val

    @classmethod
    def compile_ordered_aggregation(cls, *args: sqa.ColumnElement, order_by: list[sqa.UnaryExpression], impl):
        from sqlalchemy.dialects import postgresql
        from sqlalchemy.dialects.postgresql import aggregate_order_by

        # TODO: this is hacky but there is currently no other way except writing raw
        # SQL.
        return impl(
            *args[:-1],
            sqa.text(
                str(
                    aggregate_order_by(args[-1], *order_by).compile(
                        dialect=postgresql.psycopg2.PGDialect_psycopg2(),
                        compile_kwargs={"literal_binds": True},
                    )
                )
            ),
        )


with SqliteImpl.impl_store.impl_manager as impl:

    @impl(ops.round)
    def _round(x, decimals):
        if decimals >= 0:
            return sqa.func.ROUND(x, decimals, type_=x.type)
        # For some reason SQLite doesn't like negative decimals values
        return sqa.func.ROUND(x / (10**-decimals), type_=x.type) * (10**-decimals)

    @impl(ops.str_starts_with)
    def _str_starts_with(x, y):
        warn_non_standard(
            "SQLite: startswith is case-insensitive by default. "
            "Use the 'case_sensitive_like' pragma to change this behaviour. "
            "See https://www.sqlite.org/pragma.html#pragma_case_sensitive_like",
        )
        return x.startswith(y, autoescape=True)

    @impl(ops.str_ends_with)
    def _str_ends_with(x, y):
        warn_non_standard(
            "SQLite: endswith is case-insensitive by default. "
            "Use the 'case_sensitive_like' pragma to change this behaviour. "
            "See https://www.sqlite.org/pragma.html#pragma_case_sensitive_like",
        )
        return x.endswith(y, autoescape=True)

    @impl(ops.str_contains)
    def _str_contains(x, y, allow_regex, true_if_regex_unsupported):
        if true_if_regex_unsupported:
            return sqa.literal(True, literal_execute=True)
        warn_non_standard(
            "SQLite: contains is case-insensitive by default. "
            "Use the 'case_sensitive_like' pragma to change this behaviour. "
            "See https://www.sqlite.org/pragma.html#pragma_case_sensitive_like",
        )
        return x.contains(y, autoescape=True)

    @impl(ops.dt_millisecond)
    def _dt_millisecond(x):
        warn_non_standard(
            "SQLite returns rounded milliseconds",
        )
        frac_seconds = sqa.cast(sqa.func.STRFTIME("%f", x), sqa.Numeric())
        return sqa.cast((frac_seconds * 1e3), sqa.Integer()) % 1000

    @impl(ops.dt_microsecond)
    def _dt_microsecond(x):
        warn_non_standard(
            "SQLite returns rounded microseconds",
        )
        _1000 = sqa.literal_column("1000")
        frac_seconds = sqa.cast(sqa.func.STRFTIME("%f", x), sqa.Numeric())
        return sqa.cast(frac_seconds * 1e6, sqa.Integer()) % 1_000_000

    @impl(ops.horizontal_max)
    def _greatest(*x):
        # The SQLite MAX function returns NULL if any of the inputs are NULL
        # -> Use divide and conquer approach with coalesce to ensure correct result
        if len(x) == 1:
            return x[0]

        mid = (len(x) + 1) // 2
        left = _greatest(*x[:mid])
        right = _greatest(*x[mid:])

        return sqa.func.coalesce(sqa.func.MAX(left, right), left, right)

    @impl(ops.horizontal_min)
    def _least(*x):
        # The SQLite MIN function returns NULL if any of the inputs are NULL
        # -> Use divide and conquer approach with coalesce to ensure correct result
        if len(x) == 1:
            return x[0]

        mid = (len(x) + 1) // 2
        left = _least(*x[:mid])
        right = _least(*x[mid:])

        return sqa.func.coalesce(sqa.func.MIN(left, right), left, right)

    # TODO: we need to get the string in the right format here (so sqlite can work with
    # it)
    @impl(ops.str_to_datetime)
    def _str_to_datetime(x):
        return sqa.type_coerce(x, sqa.DateTime)

    # the SQLite floor function is cursed... it throws if you pass in a large value
    # like 1e19. surprisingly, 1e18 works... what a coincidence... :)
    @impl(ops.floor)
    def _floor(x):
        return -sqa.func.ceil(-x)

    @impl(ops.is_nan)
    def _is_nan(x):
        return False

    @impl(ops.is_not_nan)
    def _is_not_nan(x):
        return True

    @impl(ops.cbrt)
    def _cbrt(x):
        pow_impl = SqliteImpl.get_impl(ops.pow, (Float(), Float()))
        return sqa.func.sign(x) * pow_impl(sqa.func.abs(x), sqa.literal(1 / 3, type_=sqa.Double))

    @impl(ops.clip)
    def _clip(x, lower, upper):
        return sqa.func.max(sqa.func.min(x, upper), lower)

    @impl(ops.dt_day_of_week)
    def _day_of_week(x):
        return (sqa.extract("dow", x) + 6) % sqa.literal_column("7") + 1



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# This module defines the config classes provided to the user to configure
# the backend on import / export.


import sqlalchemy as sqa


class Target: ...


class Polars(Target):
    def __init__(self, *, lazy: bool = False) -> None:
        self.lazy = lazy


class Pandas(Target): ...


class DuckDb(Target): ...


class SqlAlchemy(Target):
    def __init__(self, engine: sqa.Engine, *, schema: str | None = None):
        self.engine = engine
        self.schema = schema


class Scalar(Target): ...


class Dict(Target): ...


class DictOfLists(Target): ...


class ListOfDicts(Target): ...



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import (
    NUMERIC,
    Bool,
    Date,
    Datetime,
    Duration,
    Float,
    Int,
    String,
)

add = Operator(
    "__add__",
    *(Signature(dtype, dtype, return_type=dtype) for dtype in NUMERIC),
    Signature(String(), String(), return_type=String()),
    Signature(Bool(), Bool(), return_type=Int()),
    Signature(Duration(), Duration(), return_type=Duration()),
    Signature(Datetime(), Duration(), return_type=Datetime()),
    Signature(Duration(), Datetime(), return_type=Datetime()),
    doc="Addition +",
)

sub = Operator(
    "__sub__",
    *(Signature(dtype, dtype, return_type=dtype) for dtype in NUMERIC),
    Signature(Datetime(), Datetime(), return_type=Duration()),
    Signature(Date(), Date(), return_type=Duration()),
    doc="Subtraction -",
)

mul = Operator(
    "__mul__",
    *(Signature(dtype, dtype, return_type=dtype) for dtype in NUMERIC),
    doc="Multiplication *",
)

truediv = Operator(
    "__truediv__",
    Signature(Int(), Int(), return_type=Float()),
    Signature(Float(), Float(), return_type=Float()),
    doc="True division /",
)

floordiv = Operator(
    "__floordiv__",
    Signature(Int(), Int(), return_type=Int()),
    doc="""
Integer division //

Warning
-------
The behavior of this operator differs from polars and python. Polars and python
always round towards negative infinity, whereas pydiverse.transform always
rounds towards zero, regardless of the sign. This behavior matches the one of C,
C++ and all currently supported SQL backends.

See also
--------
__mod__

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [65, -65, 65, -65],
...         "b": [7, 7, -7, -7],
...     }
... )
>>> t >> mutate(r=t.a // t.b) >> show()
shape: (4, 3)
┌─────┬─────┬─────┐
│ a   ┆ b   ┆ r   │
│ --- ┆ --- ┆ --- │
│ i64 ┆ i64 ┆ i64 │
╞═════╪═════╪═════╡
│ 65  ┆ 7   ┆ 9   │
│ -65 ┆ 7   ┆ -9  │
│ 65  ┆ -7  ┆ -9  │
│ -65 ┆ -7  ┆ 9   │
└─────┴─────┴─────┘
""",
)

mod = Operator(
    "__mod__",
    Signature(Int(), Int(), return_type=Int()),
    doc="""
The remainder of integer division %

Warning
-------
This operator behaves differently than in polars. There are at least two
conventions how `%` and :doc:`// <pydiverse.transform.ColExpr.__floordiv__>`
should behave  for negative inputs. We follow the one that C, C++ and all
currently supported SQL backends follow. This means that the output has the same
sign as the left hand side of the input, regardless of the right hand side.

See also
--------
__floordiv__

Examples
--------
>>> t = pdt.Table(
...     {
...         "a": [65, -65, 65, -65],
...         "b": [7, 7, -7, -7],
...     }
... )
>>> t >> mutate(r=t.a % t.b) >> show()
shape: (4, 3)
┌─────┬─────┬─────┐
│ a   ┆ b   ┆ r   │
│ --- ┆ --- ┆ --- │
│ i64 ┆ i64 ┆ i64 │
╞═════╪═════╪═════╡
│ 65  ┆ 7   ┆ 2   │
│ -65 ┆ 7   ┆ -2  │
│ 65  ┆ -7  ┆ 2   │
│ -65 ┆ -7  ┆ -2  │
└─────┴─────┴─────┘
""",
)
